from PyQt5.QtWidgets import QDialog

from .qt.edit_model import Ui_EditModel

from .define_external_forces import define_exernal_forces

class EditModelDialog(QDialog):

    def __init__(self, parent_window=None):
        super().__init__()
        self.ui = Ui_EditModel()
        self.ui.setupUi(self)

        self.parent_window = parent_window

        self.model = parent_window.simulation.model # indispensable pour ne garder que les modifs validées
        self.simulation = parent_window.simulation

        self.ui.enableXPeriodicBoudariesCheckBox.setChecked(self.model.x_periodic)
        self.ui.enableYPeriodicBoudariesCheckBox.setChecked(self.model.y_periodic)

        self.external_forces = define_exernal_forces(simulation=parent_window.simulation, until=parent_window.ui.iterationsSpinBox.value())
        self.ui.verticalLayout.addWidget(self.external_forces)

        self.ui.buttonBox.accepted.connect(self.accept)
        self.ui.buttonBox.rejected.connect(self.reject)


    def accept(self):
        self.model.x_periodic = self.ui.enableXPeriodicBoudariesCheckBox.isChecked()
        self.model.y_periodic = self.ui.enableYPeriodicBoudariesCheckBox.isChecked()
        self.model.params["up_apply_force_x"] = self.external_forces.up_apply_x
        self.model.params["up_apply_force_y"] = self.external_forces.up_apply_y
        self.model.params["low_block"] = self.external_forces.low_block
        self.simulation.set_Fx_ramps(*self.external_forces.up_x_component)
        self.simulation.set_Fy_ramps(*self.external_forces.up_y_component)
        super().accept()


"""Tests for Bundle.

"""

import numpy as np
import pytest

import datreant as dtr


def do_stuff(treant):
    return treant.name


def return_nothing(treant):
    b = treant.name


class CollectionsTests(object):
    """Mixin tests for collections"""

    def test_parents(self, collection, tmpdir):
        with tmpdir.as_cwd():
            t1 = dtr.Treant('free-associate/lark')
            t2 = dtr.Treant('free-associate/hark')
            t3 = dtr.Treant('characters/linus')

        col = collection(t1, t2, t3)

        assert len(col.parents()) == 2
        assert 'free-associate' in col.parents().names
        assert 'linus' not in col.parents().names
        assert 'characters' in col.parents().names

    class TestGetitem(object):
        @pytest.mark.parametrize('slx', (
            [1, 2],
            np.array([1, 2]),
        ))
        def test_fancy_index(self, filled_collection, slx):
            b, (t1, t2, t3) = filled_collection
            sl = b[slx]
            assert len(sl) == 2
            assert t2 == sl[0]
            assert t3 == sl[1]

        @pytest.mark.parametrize('slx', (
            [False, False, True],
            np.array([False, False, True]),
        ))
        def test_boolean_index(self, filled_collection, slx):
            b, (t1, t2, t3) = filled_collection
            sl = b[slx]
            assert len(sl) == 1
            assert t3 == sl[0]

        @pytest.mark.parametrize('slx', (
            slice(0, 1, None),
            slice(1, None, None),
            slice(None, None, -1),
            slice(None, None, 2),
        ))
        def test_slice_index(self, filled_collection, slx):
            b, ts = filled_collection
            sl = b[slx]
            ref = ts[slx]
            for x, y in zip(sl, ref):
                assert x == y

        def test_getitem_IE(self, filled_collection):
            bundle = filled_collection[0]
            with pytest.raises(IndexError):
                bundle[4.0]

    class TestSetOperations(object):
        def test_sub_single(self, filled_collection):
            b, (t1, t2, t3) = filled_collection
            b1 = b[[0, 1, 2]]
            b2 = b[1]
            b3 = b1 - b2
            assert len(b3) == 2
            assert t1 in b3
            assert t2 not in b3
            assert t3 in b3

        def test_sub_many(self, filled_collection):
            b, (t1, t2, t3) = filled_collection
            b1 = b[[0, 1]]
            b2 = b[[1, 2]]
            b3 = b1 - b2
            assert len(b3) == 1
            assert t1 in b3
            assert t2 not in b3

        def test_or(self, filled_collection):
            b, (t1, t2, t3) = filled_collection
            b1 = b[[0, 1]]
            b2 = b[[1, 2]]
            b3 = b1 | b2
            assert t1 in b3
            assert t2 in b3
            assert t3 in b3

        def test_and(self, filled_collection):
            b, (t1, t2, t3) = filled_collection
            b1 = b[[0, 1]]
            b2 = b[[1, 2]]
            b3 = b1 & b2
            assert t1 not in b3

        def test_xor(self, filled_collection):
            b, (t1, t2, t3) = filled_collection
            b1 = b[[0, 1]]
            b2 = b[[1, 2]]
            b3 = b1 ^ b2
            assert len(b3) == 2
            assert t1 in b3
            assert t2 not in b3
            assert t3 in b3

        def test_sub_TypeError(self, filled_collection):
            b = filled_collection[0]
            with pytest.raises(TypeError):
                b - ['this']

        def test_or_TypeError(self, filled_collection):
            b = filled_collection[0]
            with pytest.raises(TypeError):
                b | ['this']

        def test_and_TypeError(self, filled_collection):
            b = filled_collection[0]
            with pytest.raises(TypeError):
                b & ['this']

        def test_xor_TypeError(self, filled_collection):
            b = filled_collection[0]
            with pytest.raises(TypeError):
                b ^ ['this']

    class TestAddition(object):
        def test_add_many(self, filled_collection):
            b, (t1, t2, t3) = filled_collection
            b1 = b[[0, 1]]
            b2 = b[[1, 2]]
            b3 = b1 + b2
            assert len(b3) == 3
            assert t1 in b3
            assert t2 in b3
            assert t3 in b3

        def test_add_singular(self, filled_collection):
            b, (t1, t2, t3) = filled_collection
            b1 = b[[0, 1]]
            b2 = b[2]
            b3 = b1 + b2
            assert len(b3) == 3
            assert t1 in b3
            assert t2 in b3
            assert t3 in b3

        def test_add(self, filled_collection):
            b = filled_collection[0]
            with pytest.raises(TypeError):
                b + 25


class TestView(CollectionsTests):
    """Tests for Views"""

    @pytest.fixture
    def collection(self):
        return dtr.View

    @pytest.fixture
    def filled_collection(self, tmpdir):
        # returns (a bundle of [t1, t2, t3], then individal references to each)
        with tmpdir.as_cwd():
            t1 = dtr.Tree('larry')
            t2 = dtr.Leaf('curly')
            t3 = dtr.Treant('moe')
            b = dtr.View(t1, t2, t3)
            return b, (t1, t2, t3)

    class TestGetitem(CollectionsTests.TestGetitem):
        def test_getitem_name_string(self, filled_collection):
            b, (t1, t2, t3) = filled_collection
            n = t1.name

            b_new = b[n]
            assert isinstance(b_new, dtr.View)
            assert b_new[0] == t1

    class TestAddition(CollectionsTests.TestAddition):
        def test_tree_addition(self, filled_collection):
            b, (t1, t2, t3) = filled_collection
            b1 = b[[1, 2]]
            b3 = b1 + t1
            assert len(b3) == 3
            assert isinstance(b3, dtr.View)
            assert t1 in b3

        def test_leaf_addition(self, filled_collection):
            b, (t1, t2, t3) = filled_collection
            b1 = b[[0, 2]]
            b3 = b1 + t2
            assert len(b3) == 3
            assert isinstance(b3, dtr.View)
            assert t2 in b3

        def test_treant_addition(self, filled_collection):
            b, (t1, t2, t3) = filled_collection
            b1 = b[[0, 1]]
            b3 = b1 + t3
            assert len(b3) == 3
            assert isinstance(b3, dtr.View)
            assert t2 in b3

    def test_exists(self, collection, tmpdir):
        pass


class TestBundle(CollectionsTests):
    """Tests for elements of Bundle"""

    @pytest.fixture
    def collection(self):
        return dtr.Bundle

    @pytest.fixture
    def filled_collection(self, tmpdir):
        # returns (a bundle of [t1, t2, t3], then individal references to each)
        with tmpdir.as_cwd():
            t1 = dtr.Treant('larry')
            t2 = dtr.Treant('curly')
            t3 = dtr.Treant('moe')
            b = dtr.Bundle((t1, t2, t3))
            return b, (t1, t2, t3)

    @pytest.fixture
    def testtreant(self, tmpdir, request):
        with tmpdir.as_cwd():
            t = dtr.Treant('dummytreant')
        return t

    @pytest.fixture
    def testtreant2(self, tmpdir, request):
        with tmpdir.as_cwd():
            t = dtr.Treant('dummytreant2')
        return t

    def test_additive(self, tmpdir, testtreant, testtreant2, collection):
        """Test that addition of treants and collections give Bundles.

        """
        with tmpdir.as_cwd():
            assert isinstance(testtreant + testtreant2, dtr.Bundle)
            assert len(testtreant + testtreant2) == 2

            b = collection() + testtreant + testtreant2

            # beating a dead horse
            assert len(b) == 2

    class TestGetitem(CollectionsTests.TestGetitem):
        def test_getitem_name_string(self, filled_collection):
            b, (t1, t2, t3) = filled_collection
            n = t1.name

            b_new = b[n]
            assert isinstance(b_new, dtr.Bundle)
            assert b_new[0] == t1

        def test_getitem_string_KeyError(self, filled_collection):
            b = filled_collection[0]
            with pytest.raises(KeyError):
                b['not there']

    def test_get_members(self, collection, tmpdir):
        """Access members with indexing and slicing"""
        with tmpdir.as_cwd():
            t1 = dtr.Treant('larry')
            t2 = dtr.Treant('curly')
            t3 = dtr.Treant('moe')

            col = collection([[[t1, [t2, [t3]]]]])

            assert col[1] == t2

            t4 = dtr.treants.Treant('shemp')
            col = col + t4

            for member in (t1, t2, t3):
                assert member in col[:3]

            assert t4 not in col[:3]
            assert t4 == col[-1]

    def test_member_attributes(self, collection, tmpdir):
        """Get member names and abspaths"""
        with tmpdir.as_cwd():
            t1 = dtr.Treant('bigger')
            t2 = dtr.Treant('faster')
            t3 = dtr.Treant('stronger')

        col = collection(t1, t2, t3)

        names = [treant.name for treant in [t1, t2, t3]]
        assert col.names == names

        abspaths = [treant.abspath for treant in [t1, t2, t3]]
        assert col.abspaths == abspaths

    def test_map(self, collection, tmpdir):
        with tmpdir.as_cwd():
            t1 = dtr.Treant('lark')
            t2 = dtr.Treant('hark')
            t3 = dtr.Treant('linus')

        col = collection(t1, t2, t3)

        comp = [cont.name for cont in col]
        assert col.map(do_stuff) == comp
        assert col.map(do_stuff, processes=2) == comp

        assert col.map(return_nothing) is None
        assert col.map(return_nothing, processes=2) is None

    class TestGet:
        @pytest.fixture
        def filled_bundle(self, tmpdir):
            with tmpdir.as_cwd():
                t1 = dtr.Treant('one')
                t2 = dtr.Treant('two')
                t3 = dtr.Treant('three')
                t4 = dtr.Treant('four')

                t1.tags.add('odd', 'one')
                t2.tags.add('even', 'two')
                t3.tags.add('odd', 'three')
                t4.tags.add('even', 'four')
                t1.categories.add(is_even=False, value=1)
                t2.categories.add(is_even=True, value=2)
                t3.categories.add(is_even=False, value=3)
                t4.categories.add(is_even=True, value=4)

                return [t1, t2, t3, t4], dtr.Bundle([t1, t2, t3, t4])

        def test_get_blank(self, filled_bundle):
            ref, b = filled_bundle

            new = b.get()

            assert new == b

        def test_get_cats(self, filled_bundle):
            ref, b = filled_bundle

            new = b.get(is_even=True)

            assert len(new) == 2
            assert ref[1] in new
            assert ref[3] in new

        def test_get_cats_double(self, filled_bundle):
            ref, b = filled_bundle

            new = b.get(is_even=True, value=4)

            assert len(new) == 1
            assert ref[3] in new

        def test_get_cats_empty(self, filled_bundle):
            ref, b = filled_bundle

            new = b.get(is_even=True, value=1)

            assert len(new) == 0

        def test_get_cats_KeyError(self, filled_bundle):
            ref, b = filled_bundle

            new = b.get(colour='yellow')

            assert len(new) == 0

        def test_get_tags(self, filled_bundle):
            ref, b = filled_bundle

            new = b.get('odd')

            assert len(new) == 2
            assert ref[0] in new
            assert ref[2] in new

        def test_get_tags_many(self, filled_bundle):
            ref, b = filled_bundle

            new = b.get('odd', 'three')

            assert len(new) == 1
            assert ref[2] in new

        def test_get_tags_empty(self, filled_bundle):
            ref, b = filled_bundle

            new = b.get('magical')

            assert len(new) == 0

        def test_get_cat_and_tag(self, filled_bundle):
            ref, b = filled_bundle

            new = b.get('odd', value=3)

            assert len(new) == 1
            assert ref[2] in new

    class TestAggTags:
        """Test behavior of manipulating tags collectively.

        """
        def test_add_tags(self, collection, testtreant, testtreant2, tmpdir):
            with tmpdir.as_cwd():
                col = collection(testtreant, testtreant2)

                assert len(col.tags) == 0

                col.tags.add('broiled', 'not baked')

                assert len(col.tags) == 2
                for tag in ('broiled', 'not baked'):
                    assert tag in col.tags

        def test_tags_setting(self, collection, testtreant,
                              testtreant2, tmpdir):
            with tmpdir.as_cwd():
                col = collection(testtreant, testtreant2)

                assert len(col.tags) == 0

                # add as list
                col.tags = ['broiled', 'not baked']

                assert len(col.tags) == 2
                for tag in ('broiled', 'not baked'):
                    assert tag in col.tags

                col.tags.clear()

                # add as set
                col.tags = {'broiled', 'not baked'}

                assert len(col.tags) == 2
                for tag in ('broiled', 'not baked'):
                    assert tag in col.tags

                col.tags.clear()

                # add as Tags
                t = dtr.Treant('mark twain')
                t.tags.add('literature', 'quotables')
                col.tags = t.tags

                assert len(col.tags) == 2
                for tag in ('literature', 'quotables'):
                    assert tag in col.tags

        def test_tags_all(self, collection, tmpdir):
            with tmpdir.as_cwd():

                moe = dtr.Treant('moe',
                                 tags=['smartest', 'mean', 'stooge'])
                larry = dtr.Treant('larry',
                                   tags=['weird', 'means well', 'stooge'])
                curly = dtr.Treant('curly',
                                   tags=['dumb', 'nyuk-nyuk', 'stooge'])

                col = collection(moe, larry, curly)

                assert len(col.tags.all) == 1
                assert 'stooge' in col.tags.all

        def test_tags_any(self, collection, testtreant, testtreant2, tmpdir):
            with tmpdir.as_cwd():

                moe = dtr.Treant('moe',
                                 tags=['smartest', 'mean', 'stooge'])
                larry = dtr.Treant('larry',
                                   tags=['weird', 'means well', 'stooge'])
                curly = dtr.Treant('curly',
                                   tags=['dumb', 'nyuk-nyuk', 'stooge'])

                col = collection(moe, larry, curly)

                assert len(col.tags.any) == 7
                for tag in ('smartest', 'mean', 'weird', 'means well',
                            'dumb', 'nyuk-nyuk', 'stooge'):
                    assert tag in col.tags.any

        def test_tags_set_behavior(self, collection, tmpdir):
            with tmpdir.as_cwd():

                t1 = dtr.Treant('maple')
                t2 = dtr.Treant('pine')
                t3 = dtr.Treant('juniper')
                t1.tags.add({'tree', 'new jersey', 'deciduous'})
                t2.tags.add({'tree', 'new york', 'evergreen'})
                t3.tags.add({'shrub', 'new york', 'evergreen'})
                col = collection(t1, t2, t3)
                trees = dtr.Bundle('maple', 'pine')
                evergreens = dtr.Bundle('pine', 'juniper')
                tags = col.tags

                assert len(tags.any) == 6

                # test equality: __eq__ (==)
                assert t1.tags == {'tree', 'new jersey', 'deciduous'}
                assert t2.tags == {'tree', 'new york', 'evergreen'}
                assert t3.tags == {'shrub', 'new york', 'evergreen'}

                # test subset: __lt__ (<)
                assert not t1.tags < {'tree', 'new jersey', 'deciduous'}
                assert tags < {'tree', 'new jersey', 'deciduous'}
                assert t1.tags < tags.any
                assert t2.tags < tags.any
                assert t3.tags < tags.any

                # test difference: __sub__ (-)
                assert t1.tags - {'tree'} == {'new jersey', 'deciduous'}
                assert trees.tags - {'tree'} == set()

                # test right difference: __rsub__ (-)
                evergreen_ny_shrub = {'evergreen', 'new york', 'shrub'}
                dec_nj_sh = {'deciduous', 'new jersey', 'shrub'}
                assert tags.any - t1.tags == evergreen_ny_shrub
                assert tags.any - evergreens.tags - trees.tags == dec_nj_sh
                assert {'tree'} - trees.tags == set()

                # test union: __or__ (|)
                evergreen_ny_shrub = {'evergreen', 'new york', 'shrub'}
                assert evergreens.tags | t3.tags == evergreen_ny_shrub
                assert t1.tags | t2.tags | t3.tags == tags.any

                # test right union: __ror__ (|)
                assert {'shrub'} | evergreens.tags == evergreen_ny_shrub
                assert t3.tags | {'tree'} == {'tree'} | t3.tags

                # test intersection: __and__ (&)
                evergreen_ny = {'evergreen', 'new york'}
                assert evergreens.tags & t3.tags == evergreen_ny
                assert t1.tags & t2.tags & t3.tags == tags.all

                # test right intersection: __rand__ (&)
                assert evergreen_ny_shrub & evergreens.tags == evergreen_ny
                assert t3.tags & {'shrub'} == {'shrub'} & t3.tags

                # test symmetric difference: __xor__ (^)
                evergreen_ny_tree = {'evergreen', 'new york', 'tree'}
                assert trees.tags ^ evergreens.tags == evergreen_ny_tree
                assert evergreens.tags ^ t3.tags == {'shrub'}
                assert t1.tags ^ t2.tags ^ t3.tags == dec_nj_sh

                # test right symmetric difference: __rxor__ (^)
                assert {'new york'} ^ evergreens.tags == {'evergreen'}
                assert {'shrub'} ^ trees.tags == t2.tags ^ t3.tags

                # type_error_msg = "Operands must be AggTags, Tags, or a set."
                with pytest.raises(TypeError) as e:
                    ['tree'] == trees.tags
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    trees.tags < ['tree']
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    ['tree'] - trees.tags
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    trees.tags - ['tree']
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    ['tree'] | trees.tags
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    trees.tags | ['tree']
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    ['tree'] & trees.tags
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    trees.tags & ['tree']
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    ['tree'] ^ trees.tags
                # assert e.value.message == type_error_msg
                with pytest.raises(TypeError) as e:
                    trees.tags ^ ['tree']
                # assert e.value.message == type_error_msg

        def test_tags_getitem(self, collection, testtreant,
                              testtreant2, tmpdir):
            with tmpdir.as_cwd():

                t1 = dtr.Treant('maple')
                t2 = dtr.Treant('pine')
                t1.tags.add({'tree', 'new jersey', 'deciduous'})
                t2.tags.add({'tree', 'new york', 'evergreen'})
                col = collection(t1, t2)
                tags = col.tags

                assert len(tags.any) == 5
                # test single tags
                assert tags['tree'] == [True, True]
                assert tags['deciduous'] == [True, False]
                assert tags['evergreen'] == [False, True]
                assert tags['new jersey'] == [True, False]
                assert tags['new york'] == [False, True]
                assert tags[{'tree'}] == [False, False]
                assert tags[{'deciduous'}] == [False, True]
                assert tags[{'evergreen'}] == [True, False]
                assert tags[{'new jersey'}] == [False, True]
                assert tags[{'new york'}] == [True, False]

                # test for Treants with ALL the given tags
                assert tags[['tree', 'deciduous']] == [True, False]
                assert tags[['tree', 'evergreen']] == [False, True]
                assert tags[['new jersey', 'evergreen']] == [False, False]

                # test for Treants with ANY of the given tags
                assert tags[('tree', 'deciduous')] == [True, True]
                assert tags[('deciduous', 'evergreen')] == [True, True]
                assert tags[('new york', 'evergreen')] == [False, True]

                # test for Treants without at least one of the given tags
                assert tags[{'deciduous', 'evergreen'}] == [True, True]
                assert tags[{'tree', 'deciduous'}] == [False, True]
                assert tags[{'tree', 'new york', 'evergreen'}] == [True, False]

                # complex logic tests

                # give those that are evergreen or in NY AND also not deciduous
                selection = [('evergreen', 'new york'), {'deciduous'}]
                assert tags[selection] == [False, True]
                # give those that are evergreen or in NY AND also not a tree
                selection = [('evergreen', 'new york'), {'tree'}]
                assert tags[selection] == [False, False]
                # give a tree that's in NJ OR anything that's not evergreen
                selection = (['tree', 'new jersey'], {'evergreen'})
                assert tags[selection] == [True, False]
                # cannot be a tree in NJ, AND must also be deciduous
                # I.e., give all deciduous things that aren't trees in NJ
                selection = [{'tree', 'new jersey'}, 'deciduous']
                assert tags[selection] == [False, False]

        def test_tags_fuzzy(self, collection, testtreant, testtreant2, tmpdir):
            with tmpdir.as_cwd():

                t1 = dtr.Treant('maple')
                t2 = dtr.Treant('pine')
                t1.tags.add({'tree', 'new jersey', 'deciduous'})
                t2.tags.add({'tree', 'new york', 'evergreen'})
                col = collection(t1, t2)
                tags = col.tags

                assert len(tags.any) == 5

                all_tree1 = tags.fuzzy('tree', threshold=80, scope='all')
                all_tree2 = tags.fuzzy('tree')
                assert all_tree1 == all_tree2
                assert all_tree2 == ('tree',)

                any_deciduous = tags.fuzzy('deciduous', scope='any')
                assert any_deciduous == ('deciduous',)
                all_evergreen = tags.fuzzy('evergreen')
                assert all_evergreen == ()

                # check that fuzzy matching is independent of threshold when
                # exact tag is present in all members
                all_tree_strict = tags.fuzzy('tree', threshold=99)
                assert all_tree_strict == ('tree',)
                all_tree_tolerant = tags.fuzzy('tree', threshold=0)
                assert all_tree_tolerant == ('tree',)

                # check that fuzzy matching will give differing tags when
                # members have similar tag names ('new') and the threshold is
                # varied
                all_ny = tags.fuzzy('new york')
                assert all_ny == ()
                any_ny_strict = tags.fuzzy('new york', scope='any')
                assert any_ny_strict == ('new york',)
                any_ny_tol = tags.fuzzy('new york', threshold=50, scope='any')
                assert set(any_ny_tol) == {'new york', 'new jersey'}

                # check fuzzy matching for multiple tags (scope='all')
                new_ever = ['new', 'evergreen']
                all_mul_strict = tags.fuzzy(new_ever, threshold=80)
                assert all_mul_strict == ()
                all_mul_tol = tags.fuzzy(new_ever, threshold=30)
                assert all_mul_tol == ('tree',)

                # check fuzzy matching for multiple tags (scope='any')
                new_tree = ['new', 'tree']
                any_mul_stric = tags.fuzzy(new_tree, threshold=90, scope='any')
                assert any_mul_stric == ('tree',)
                any_mul_tol = tags.fuzzy(new_tree, threshold=80, scope='any')
                assert set(any_mul_tol) == {'new york', 'new jersey', 'tree'}
                nj_decid = ['new jersey', 'decid']
                any_mul_njdec = tags.fuzzy(nj_decid, threshold=80, scope='any')
                assert set(any_mul_njdec) == {'new jersey', 'deciduous'}

        def test_tags_filter(self, collection, testtreant,
                             testtreant2, tmpdir):
            with tmpdir.as_cwd():

                maple = dtr.Treant('maple')
                pine = dtr.Treant('pine')
                maple.tags.add({'tree', 'new jersey', 'deciduous'})
                pine.tags.add({'tree', 'new york', 'evergreen'})
                col = collection(maple, pine)
                tags = col.tags

                maple_bund = dtr.Bundle(maple)
                pine_bund = dtr.Bundle(pine)

                assert len(tags.any) == 5

                # filter using single tags
                assert tags.filter('tree') == col
                assert tags.filter({'tree'}) == dtr.Bundle()
                assert tags.filter('deciduous') == maple_bund
                assert tags.filter('evergreen') == pine_bund
                assert tags.filter('new jersey') == maple_bund
                assert tags.filter('new york') == pine_bund

                # filter Treants that DON'T have a given tag
                assert tags.filter({'new york'}) == maple_bund
                assert tags.filter({'deciduous'}) == pine_bund

                # filter Treants containing all of the tags
                assert tags.filter(['deciduous', 'tree']) == maple_bund
                assert tags.filter(['evergreen', 'tree']) == pine_bund
                assert tags.filter(['deciduous', 'new york']) == dtr.Bundle()

                # filter Treants containing any of the tags tags
                assert tags.filter(('evergreen', 'tree')) == col
                assert tags.filter(('deciduous', 'new york')) == col
                assert tags.filter(('evergreen', 'new york')) == pine_bund

                # filter Treants that exclude any of the provided tags
                assert tags.filter({'deciduous', 'new york'}) == col
                assert tags.filter({'deciduous', 'new jersey'}) == pine_bund
                assert tags.filter({'evergreen', 'tree'}) == maple_bund

                # complex logic tests

                # give those that are evergreen or in NY AND also not deciduous
                selection = [('evergreen', 'new york'), {'deciduous'}]
                assert tags.filter(selection) == pine_bund
                # give those that are evergreen or in NY AND also not a tree
                selection = [('evergreen', 'new york'), {'tree'}]
                assert tags.filter(selection) == dtr.Bundle()
                # give a tree that's in NJ OR anything that's not evergreen
                selection = (['tree', 'new jersey'], {'evergreen'})
                assert tags.filter(selection) == maple_bund
                # cannot be a tree in NJ, AND must also be deciduous
                # I.e., give all deciduous things that aren't trees in NJ
                selection = [{'tree', 'new jersey'}, 'deciduous']
                assert tags.filter(selection) == dtr.Bundle()

    class TestAggCategories:
        """Test behavior of manipulating categories collectively.

        """
        def test_add_categories(self, collection, testtreant, testtreant2,
                                tmpdir):
            with tmpdir.as_cwd():
                # add a couple test Treants to collection
                col = collection(testtreant, testtreant2)
                assert len(col.categories) == 0

                # add 'age' and 'bark' as categories of this collection
                col.categories.add({'age': 42}, bark='smooth')
                assert len(col.categories) == 2

                for member in col:
                    assert member.categories['age'] == 42
                    assert member.categories['bark'] == 'smooth'
                for key in ['age', 'bark']:
                    assert key in col.categories.any

                t1 = dtr.Treant('hickory')
                t1.categories.add(bark='shaggy', species='ovata')
                col += collection(t1)
                assert len(col.categories) == 1
                assert len(col.categories.all) == 1
                assert len(col.categories.any) == 3

                col.categories.add(location='USA')
                assert len(col.categories) == 2
                assert len(col.categories.all) == 2
                assert len(col.categories.any) == 4
                for member in col:
                    assert member.categories['location'] == 'USA'

        def test_categories_getitem(self, collection, testtreant, testtreant2,
                                    tmpdir):
            with tmpdir.as_cwd():
                # add a couple test Treants to collection
                col = collection((testtreant, testtreant2))

                # add 'age' and 'bark' as categories of this collection
                col.categories.add({'age': 42, 'bark': 'smooth'})

                t1 = dtr.Treant('maple')
                t2 = dtr.Treant('sequoia')
                t1.categories.add({'age': 'seedling', 'bark': 'rough',
                                   'type': 'deciduous'})
                t2.categories.add({'age': 'adult', 'bark': 'rough',
                                   'type': 'evergreen', 'nickname': 'redwood'})

                col += collection((t1, t2))
                assert len(col.categories) == 2
                assert len(col.categories.any) == 4

                # test values for each category in the collection
                age_list = [42, 42, 'seedling', 'adult']
                assert age_list == col.categories['age']
                bark_list = ['smooth', 'smooth', 'rough', 'rough']
                assert bark_list == col.categories['bark']
                type_list = [None, None, 'deciduous', 'evergreen']
                assert type_list == col.categories['type']
                nick_list = [None, None,  None, 'redwood']
                assert nick_list == col.categories['nickname']

                # test list of keys as input
                cat_list = [age_list, type_list]
                assert cat_list == col.categories[['age', 'type']]

                # test set of keys as input
                cat_set = {'bark': bark_list, 'nickname': nick_list}
                assert cat_set == col.categories[{'bark', 'nickname'}]

        def test_categories_setitem(self, collection, testtreant, testtreant2,
                                    tmpdir):
            with tmpdir.as_cwd():
                # add a couple test Treants to collection
                col = collection(testtreant, testtreant2)
                # add 'age' and 'bark' as categories of this collection
                col.categories.add({'age': 42, 'bark': 'smooth'})

                t1 = dtr.Treant('maple')
                t2 = dtr.Treant('sequoia')
                t1.categories.add({'age': 'seedling', 'bark': 'rough',
                                   'type': 'deciduous'})
                t2.categories.add({'age': 'adult', 'bark': 'rough',
                                   'type': 'evergreen', 'nickname': 'redwood'})
                col += collection(t1, t2)

                # test setting a category when all members have it
                for value in col.categories['age']:
                    assert value in [42, 42, 'seedling', 'adult']
                col.categories['age'] = 'old'
                for value in col.categories['age']:
                    assert value in ['old', 'old', 'old', 'old']

                # test setting a new category (no members have it)
                assert 'location' not in col.categories.any
                col.categories['location'] = 'USA'
                for value in col.categories['location']:
                    assert value in ['USA', 'USA', 'USA', 'USA']

                # test setting a category that only some members have
                assert 'nickname' in col.categories.any
                assert 'nickname' not in col.categories.all
                col.categories['nickname'] = 'friend'
                for value in col.categories['nickname']:
                    assert value in ['friend', 'friend', 'friend', 'friend']

                # test setting values for individual members
                assert 'favorite ice cream' not in col.categories
                ice_creams = ['rocky road',
                              'americone dream',
                              'moose tracks',
                              'vanilla']
                col.categories['favorite ice cream'] = ice_creams

                for member, ice_cream in zip(col, ice_creams):
                    assert member.categories['favorite ice cream'] == ice_cream

        def test_categories_all(self, collection, testtreant, testtreant2,
                                tmpdir):
            with tmpdir.as_cwd():
                # add a couple test Treants to collection
                col = collection(testtreant, testtreant2)
                # add 'age' and 'bark' as categories of this collection
                col.categories.add({'age': 42}, bark='bare')

                # add categories to 'hickory' Treant, then add to collection
                t1 = dtr.Treant('hickory')
                t1.categories.add(bark='shaggy', species='ovata')
                col += collection(t1)
                # check the contents of 'bark', ensure 'age' and 'species' are
                # not shared categories of the collection
                col += collection(t1)
                common_categories = col.categories.all
                assert len(col.categories) == len(common_categories)
                assert 'age' not in common_categories
                assert 'species' not in common_categories
                assert common_categories['bark'] == ['bare', 'bare', 'shaggy']

                # add 'location' category to collection
                col.categories.add(location='USA')
                common_categories = col.categories.all
                # ensure all members have 'USA' for their 'location'
                assert len(col.categories) == len(common_categories)
                assert 'age' not in common_categories
                assert 'species' not in common_categories
                assert common_categories['bark'] == ['bare', 'bare', 'shaggy']
                assert common_categories['location'] == ['USA', 'USA', 'USA']

                # add 'location' category to collection
                col.categories.remove('bark')
                common_categories = col.categories.all
                # check that only 'location' is a shared category
                assert len(col.categories) == len(common_categories)
                assert 'age' not in common_categories
                assert 'bark' not in common_categories
                assert 'species' not in common_categories
                assert common_categories['location'] == ['USA', 'USA', 'USA']

        def test_categories_any(self, collection, testtreant, testtreant2,
                                tmpdir):
            with tmpdir.as_cwd():
                # add a couple test Treants to collection
                col = collection(testtreant, testtreant2)
                # add 'age' and 'bark' as categories of this collection
                col.categories.add({'age': 42}, bark='smooth')
                assert len(col.categories.any) == 2

                # add categories to 'hickory' Treant, then add to collection
                t1 = dtr.Treant('hickory')
                t1.categories.add(bark='shaggy', species='ovata')
                col += collection(t1)
                # check the contents of 'bark', ensure 'age' and 'species' are
                # not shared categories of the collection
                every_category = col.categories.any
                assert len(every_category) == 3
                assert every_category['age'] == [42, 42, None]
                assert every_category['bark'] == ['smooth', 'smooth', 'shaggy']
                assert every_category['species'] == [None, None, 'ovata']

                # add 'location' category to collection
                col.categories.add(location='USA')
                every_category = col.categories.any
                # ensure all members have 'USA' for their 'location'
                assert len(every_category) == 4
                assert every_category['age'] == [42, 42, None]
                assert every_category['bark'] == ['smooth', 'smooth', 'shaggy']
                assert every_category['species'] == [None, None, 'ovata']
                assert every_category['location'] == ['USA', 'USA', 'USA']

                # add 'sprout' to 'age' category of 'hickory' Treant
                t1.categories['age'] = 'sprout'
                every_category = col.categories.any
                # check 'age' is category for 'hickory' and is 'sprout'
                assert len(every_category) == 4
                assert every_category['age'] == [42, 42, 'sprout']
                assert every_category['bark'] == ['smooth', 'smooth', 'shaggy']
                assert every_category['species'] == [None, None, 'ovata']
                assert every_category['location'] == ['USA', 'USA', 'USA']

                # add 'location' category to collection
                col.categories.remove('bark')
                every_category = col.categories.any
                # check that only 'location' is a shared category
                assert len(every_category) == 3
                assert every_category['age'] == [42, 42, 'sprout']
                assert every_category['species'] == [None, None, 'ovata']
                assert every_category['location'] == ['USA', 'USA', 'USA']
                assert 'bark' not in every_category

        def test_categories_remove(self, collection, testtreant, testtreant2,
                                   tmpdir):
            with tmpdir.as_cwd():
                t1 = dtr.Treant('maple')
                t2 = dtr.Treant('sequoia')

                col = collection(t1, t2)
                col.categories.add({'age': 'sprout'}, bark='rough')

                # add a couple test Treants to collection
                col += collection(testtreant, testtreant2)
                assert len(col.categories) == 0
                assert len(col.categories.any) == 2

                # add 'USA', ensure 'location', 'age', 'bark' is a category in
                # at least one of the members
                col.categories.add(location='USA')
                assert len(col.categories) == 1
                for key in ['location', 'age', 'bark']:
                    assert key in col.categories.any
                # ensure 'age' and 'bark' are each not categories for all
                # members in collection
                assert 'age' not in col.categories
                assert 'bark' not in col.categories

                # remove 'bark', test for any instance of 'bark' in the
                # collection
                col.categories.remove('bark')
                assert len(col.categories) == 1
                for key in ['location', 'age']:
                    assert key in col.categories.any
                assert 'bark' not in col.categories.any

                # remove 'age', test that 'age' is not a category for any
                # member in collection
                col.categories.remove('age')
                for member in col:
                    assert 'age' not in member.categories
                # test that 'age' is not a category of this collection
                assert 'age' not in col.categories.any

        def test_categories_keys(self, collection, testtreant, testtreant2,
                                 tmpdir):
            with tmpdir.as_cwd():
                # add a couple test Treants to collection
                col = collection(testtreant, testtreant2)
                col.categories.add({'age': 42, 'bark': 'smooth'})

                t1 = dtr.Treant('maple')
                t2 = dtr.Treant('sequoia')
                t1.categories.add({'age': 'seedling', 'bark': 'rough',
                                   'type': 'deciduous'})
                t2.categories.add({'age': 'adult', 'bark': 'rough',
                                   'type': 'evergreen', 'nickname': 'redwood'})
                col += collection(t1, t2)

                for k in col.categories.keys(scope='all'):
                    for member in col:
                        assert k in member.categories

                for k in col.categories.keys(scope='any'):
                    for member in col:
                        if k == 'nickname':
                            if member.name == 'maple':
                                assert k not in member.categories
                            elif member.name == 'sequoia':
                                assert k in member.categories
                        elif k == 'type':
                            if (member.name != 'maple' and
                                    member.name != 'sequoia'):
                                assert k not in member.categories

                        else:
                            assert k in member.categories

        def test_categories_values(self, collection, testtreant, testtreant2,
                                   tmpdir):
            with tmpdir.as_cwd():
                # add a couple test Treants to collection
                col = collection(testtreant, testtreant2)
                col.categories.add({'age': 'young', 'bark': 'smooth'})

                t1 = dtr.Treant('maple')
                t2 = dtr.Treant('sequoia')
                t1.categories.add({'age': 'seedling', 'bark': 'rough',
                                   'type': 'deciduous'})
                t2.categories.add({'age': 'adult', 'bark': 'rough',
                                   'type': 'evergreen', 'nickname': 'redwood'})
                col += collection(t1, t2)

                for scope in ('all', 'any'):
                    for i, v in enumerate(
                            col.categories.values(scope=scope)):
                        assert v == col.categories[
                                col.categories.keys(scope=scope)[i]]

        def test_categories_groupby(self, collection, testtreant, testtreant2,
                                    tmpdir):
            with tmpdir.as_cwd():
                t1 = dtr.Treant('maple')
                t2 = dtr.Treant('sequoia')
                t3 = dtr.Treant('elm')
                t4 = dtr.Treant('oak')
                t1.categories.add({'age': 'young', 'bark': 'smooth',
                                   'type': 'deciduous'})
                t2.categories.add({'age': 'adult', 'bark': 'fibrous',
                                   'type': 'evergreen', 'nickname': 'redwood'})
                t3.categories.add({'age': 'old', 'bark': 'mossy',
                                   'type': 'deciduous', 'health': 'poor'})
                t4.categories.add({'age': 'young', 'bark': 'mossy',
                                   'type': 'deciduous', 'health': 'good'})
                col = collection(t1, t2, t3, t4)

                age_group = col.categories.groupby('age')
                assert {t1, t4} == set(age_group['young'])
                assert {t2} == set(age_group['adult'])
                assert {t3} == set(age_group['old'])

                bark_group = col.categories.groupby('bark')
                assert {t1} == set(bark_group['smooth'])
                assert {t2} == set(bark_group['fibrous'])
                assert {t3, t4} == set(bark_group['mossy'])

                type_group = col.categories.groupby('type')
                assert {t1, t3, t4} == set(type_group['deciduous'])
                assert {t2} == set(type_group['evergreen'])

                nick_group = col.categories.groupby('nickname')
                assert {t2} == set(nick_group['redwood'])
                for bundle in nick_group.values():
                    assert {t1, t3, t4}.isdisjoint(set(bundle))

                health_group = col.categories.groupby('health')
                assert {t3} == set(health_group['poor'])
                assert {t4} == set(health_group['good'])
                for bundle in health_group.values():
                    assert {t1, t2}.isdisjoint(set(bundle))

                # test list of keys as input
                age_bark = col.categories.groupby(['age', 'bark'])
                assert len(age_bark) == 4
                assert {t1} == set(age_bark[('young', 'smooth')])
                assert {t2} == set(age_bark[('adult', 'fibrous')])
                assert {t3} == set(age_bark[('old', 'mossy')])
                assert {t4} == set(age_bark[('young', 'mossy')])

                type_health = col.categories.groupby(['type', 'health'])
                assert len(type_health) == 2
                assert {t3} == set(type_health[('deciduous', 'poor')])
                assert {t4} == set(type_health[('deciduous', 'good')])
                for bundle in type_health.values():
                    assert {t1, t2}.isdisjoint(set(bundle))

                type_health = col.categories.groupby(['health', 'type'])
                assert len(type_health) == 2
                assert {t3} == set(type_health[('poor', 'deciduous')])
                assert {t4} == set(type_health[('good', 'deciduous')])
                for bundle in type_health.values():
                    assert {t1, t2}.isdisjoint(set(bundle))

                age_nick = col.categories.groupby(['age', 'nickname'])
                assert len(age_nick) == 1
                assert {t2} == set(age_nick['adult', 'redwood'])
                for bundle in age_nick.values():
                    assert {t1, t3, t4}.isdisjoint(set(bundle))

                keys = ['age', 'bark', 'health']
                age_bark_health = col.categories.groupby(keys)
                assert len(age_bark_health) == 2
                assert {t3} == set(age_bark_health[('old', 'mossy', 'poor')])
                assert {t4} == set(age_bark_health[('young', 'mossy', 'good')])
                for bundle in age_bark_health.values():
                    assert {t1, t2}.isdisjoint(set(bundle))

                keys = ['age', 'bark', 'type', 'nickname']
                abtn = col.categories.groupby(keys)
                assert len(abtn) == 1
                assert {t2} == set(abtn[('adult', 'fibrous', 'evergreen',
                                         'redwood')])
                for bundle in abtn.values():
                    assert {t1, t3, t4}.isdisjoint(set(bundle))

                keys = ['bark', 'nickname', 'type', 'age']
                abtn2 = col.categories.groupby(keys)
                assert len(abtn2) == 1
                assert {t2} == set(abtn2[('fibrous', 'redwood', 'evergreen',
                                          'adult')])
                for bundle in abtn2.values():
                    assert {t1, t3, t4}.isdisjoint(set(bundle))

                keys = ['health', 'nickname']
                health_nick = col.categories.groupby(keys)
                assert len(health_nick) == 0
                for bundle in health_nick.values():
                    assert {t1, t2, t3, t4}.isdisjoint(set(bundle))

                # Test key TypeError in groupby
                with pytest.raises(TypeError) as e:
                    col.categories.groupby({'health', 'nickname'})



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'create_model.ui'
#
# Created by: PyQt5 UI code generator 5.11.3
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_CreateModel(object):
    def setupUi(self, CreateModel):
        CreateModel.setObjectName("CreateModel")
        CreateModel.resize(508, 537)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(CreateModel.sizePolicy().hasHeightForWidth())
        CreateModel.setSizePolicy(sizePolicy)
        CreateModel.setWizardStyle(QtWidgets.QWizard.ModernStyle)
        self.speciesPage = QtWidgets.QWizardPage()
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.speciesPage.sizePolicy().hasHeightForWidth())
        self.speciesPage.setSizePolicy(sizePolicy)
        self.speciesPage.setObjectName("speciesPage")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.speciesPage)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.groupBox_2 = QtWidgets.QGroupBox(self.speciesPage)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_2.sizePolicy().hasHeightForWidth())
        self.groupBox_2.setSizePolicy(sizePolicy)
        self.groupBox_2.setObjectName("groupBox_2")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox_2)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.layout_a = QtWidgets.QVBoxLayout()
        self.layout_a.setObjectName("layout_a")
        self.verticalLayout_2.addLayout(self.layout_a)
        self.horizontalLayout.addWidget(self.groupBox_2)
        self.groupBox_1 = QtWidgets.QGroupBox(self.speciesPage)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_1.sizePolicy().hasHeightForWidth())
        self.groupBox_1.setSizePolicy(sizePolicy)
        self.groupBox_1.setObjectName("groupBox_1")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox_1)
        self.verticalLayout.setObjectName("verticalLayout")
        self.layout_b = QtWidgets.QVBoxLayout()
        self.layout_b.setObjectName("layout_b")
        self.verticalLayout.addLayout(self.layout_b)
        self.horizontalLayout.addWidget(self.groupBox_1)
        CreateModel.addPage(self.speciesPage)
        self.spatialPage = QtWidgets.QWizardPage()
        self.spatialPage.setObjectName("spatialPage")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.spatialPage)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.groupBox = QtWidgets.QGroupBox(self.spatialPage)
        self.groupBox.setObjectName("groupBox")
        self.formLayout_2 = QtWidgets.QFormLayout(self.groupBox)
        self.formLayout_2.setObjectName("formLayout_2")
        self.gridWidthNumberOfAtomsLabel = QtWidgets.QLabel(self.groupBox)
        self.gridWidthNumberOfAtomsLabel.setObjectName("gridWidthNumberOfAtomsLabel")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.gridWidthNumberOfAtomsLabel)
        self.gridWidth = QtWidgets.QSpinBox(self.groupBox)
        self.gridWidth.setMaximum(100000000)
        self.gridWidth.setSingleStep(20)
        self.gridWidth.setProperty("value", 100)
        self.gridWidth.setObjectName("gridWidth")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.gridWidth)
        self.gridHeightNumberOfAtomsLabel = QtWidgets.QLabel(self.groupBox)
        self.gridHeightNumberOfAtomsLabel.setObjectName("gridHeightNumberOfAtomsLabel")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.gridHeightNumberOfAtomsLabel)
        self.gridHeight = QtWidgets.QSpinBox(self.groupBox)
        self.gridHeight.setReadOnly(True)
        self.gridHeight.setMaximum(100000000)
        self.gridHeight.setSingleStep(20)
        self.gridHeight.setProperty("value", 100)
        self.gridHeight.setObjectName("gridHeight")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.gridHeight)
        self.keepRatioLabel = QtWidgets.QLabel(self.groupBox)
        self.keepRatioLabel.setObjectName("keepRatioLabel")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.keepRatioLabel)
        self.keepRatioCheckBox = QtWidgets.QCheckBox(self.groupBox)
        self.keepRatioCheckBox.setChecked(True)
        self.keepRatioCheckBox.setObjectName("keepRatioCheckBox")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.keepRatioCheckBox)
        self.distanceBetweenAtomsLabel = QtWidgets.QLabel(self.groupBox)
        self.distanceBetweenAtomsLabel.setObjectName("distanceBetweenAtomsLabel")
        self.formLayout_2.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.distanceBetweenAtomsLabel)
        self.distanceBetweenAtoms = QtWidgets.QLineEdit(self.groupBox)
        self.distanceBetweenAtoms.setText("")
        self.distanceBetweenAtoms.setClearButtonEnabled(False)
        self.distanceBetweenAtoms.setObjectName("distanceBetweenAtoms")
        self.formLayout_2.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.distanceBetweenAtoms)
        self.label = QtWidgets.QLabel(self.groupBox)
        self.label.setObjectName("label")
        self.formLayout_2.setWidget(5, QtWidgets.QFormLayout.LabelRole, self.label)
        self.label_atom_number = QtWidgets.QLabel(self.groupBox)
        self.label_atom_number.setObjectName("label_atom_number")
        self.formLayout_2.setWidget(5, QtWidgets.QFormLayout.FieldRole, self.label_atom_number)
        self.boxWidthLabel = QtWidgets.QLabel(self.groupBox)
        self.boxWidthLabel.setObjectName("boxWidthLabel")
        self.formLayout_2.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.boxWidthLabel)
        self.boxWidthLineEdit = QtWidgets.QLineEdit(self.groupBox)
        self.boxWidthLineEdit.setObjectName("boxWidthLineEdit")
        self.formLayout_2.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.boxWidthLineEdit)
        self.verticalLayout_3.addWidget(self.groupBox)
        self.groupBox_4 = QtWidgets.QGroupBox(self.spatialPage)
        self.groupBox_4.setObjectName("groupBox_4")
        self.formLayout_3 = QtWidgets.QFormLayout(self.groupBox_4)
        self.formLayout_3.setObjectName("formLayout_3")
        self.xPeriodicBoundariesLabel = QtWidgets.QLabel(self.groupBox_4)
        self.xPeriodicBoundariesLabel.setObjectName("xPeriodicBoundariesLabel")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.xPeriodicBoundariesLabel)
        self.xPeriodicBoundariesCheckBox = QtWidgets.QCheckBox(self.groupBox_4)
        self.xPeriodicBoundariesCheckBox.setChecked(True)
        self.xPeriodicBoundariesCheckBox.setObjectName("xPeriodicBoundariesCheckBox")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.xPeriodicBoundariesCheckBox)
        self.yPeriodicBoudariesLabel = QtWidgets.QLabel(self.groupBox_4)
        self.yPeriodicBoudariesLabel.setObjectName("yPeriodicBoudariesLabel")
        self.formLayout_3.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.yPeriodicBoudariesLabel)
        self.yPeriodicBoudariesCheckBox = QtWidgets.QCheckBox(self.groupBox_4)
        self.yPeriodicBoudariesCheckBox.setChecked(True)
        self.yPeriodicBoudariesCheckBox.setObjectName("yPeriodicBoudariesCheckBox")
        self.formLayout_3.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.yPeriodicBoudariesCheckBox)
        self.verticalLayout_3.addWidget(self.groupBox_4)
        self.groupBox_3 = QtWidgets.QGroupBox(self.spatialPage)
        self.groupBox_3.setObjectName("groupBox_3")
        self.formLayout = QtWidgets.QFormLayout(self.groupBox_3)
        self.formLayout.setObjectName("formLayout")
        self.firstSpeciesMoleFractionLabel = QtWidgets.QLabel(self.groupBox_3)
        self.firstSpeciesMoleFractionLabel.setObjectName("firstSpeciesMoleFractionLabel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.firstSpeciesMoleFractionLabel)
        self.firstSpeciesMoleFraction = QtWidgets.QDoubleSpinBox(self.groupBox_3)
        self.firstSpeciesMoleFraction.setMaximum(1.0)
        self.firstSpeciesMoleFraction.setSingleStep(0.1)
        self.firstSpeciesMoleFraction.setProperty("value", 1.0)
        self.firstSpeciesMoleFraction.setObjectName("firstSpeciesMoleFraction")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.firstSpeciesMoleFraction)
        self.verticalLayout_3.addWidget(self.groupBox_3)
        self.previewButton = QtWidgets.QPushButton(self.spatialPage)
        self.previewButton.setObjectName("previewButton")
        self.verticalLayout_3.addWidget(self.previewButton)
        CreateModel.addPage(self.spatialPage)
        self.otherPage = QtWidgets.QWizardPage()
        self.otherPage.setObjectName("otherPage")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.otherPage)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.groupBox_5 = QtWidgets.QGroupBox(self.otherPage)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_5.sizePolicy().hasHeightForWidth())
        self.groupBox_5.setSizePolicy(sizePolicy)
        self.groupBox_5.setObjectName("groupBox_5")
        self.formLayout_4 = QtWidgets.QFormLayout(self.groupBox_5)
        self.formLayout_4.setObjectName("formLayout_4")
        self.temperatureKLabel = QtWidgets.QLabel(self.groupBox_5)
        self.temperatureKLabel.setObjectName("temperatureKLabel")
        self.formLayout_4.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.temperatureKLabel)
        self.temperatureKDoubleSpinBox = QtWidgets.QDoubleSpinBox(self.groupBox_5)
        self.temperatureKDoubleSpinBox.setMinimum(0.0)
        self.temperatureKDoubleSpinBox.setMaximum(100000000000000.0)
        self.temperatureKDoubleSpinBox.setProperty("value", 1.0)
        self.temperatureKDoubleSpinBox.setObjectName("temperatureKDoubleSpinBox")
        self.formLayout_4.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.temperatureKDoubleSpinBox)
        self.verticalLayout_4.addWidget(self.groupBox_5)
        self.groupBox_7 = QtWidgets.QGroupBox(self.otherPage)
        self.groupBox_7.setObjectName("groupBox_7")
        self.formLayout_6 = QtWidgets.QFormLayout(self.groupBox_7)
        self.formLayout_6.setObjectName("formLayout_6")
        self.timestepLabel = QtWidgets.QLabel(self.groupBox_7)
        self.timestepLabel.setObjectName("timestepLabel")
        self.formLayout_6.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.timestepLabel)
        self.timestepLineEdit = QtWidgets.QLineEdit(self.groupBox_7)
        self.timestepLineEdit.setObjectName("timestepLineEdit")
        self.formLayout_6.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.timestepLineEdit)
        self.resetTimestep = QtWidgets.QToolButton(self.groupBox_7)
        self.resetTimestep.setObjectName("resetTimestep")
        self.formLayout_6.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.resetTimestep)
        self.verticalLayout_4.addWidget(self.groupBox_7)
        self.groupBox_6 = QtWidgets.QGroupBox(self.otherPage)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_6.sizePolicy().hasHeightForWidth())
        self.groupBox_6.setSizePolicy(sizePolicy)
        self.groupBox_6.setObjectName("groupBox_6")
        self.formLayout_5 = QtWidgets.QFormLayout(self.groupBox_6)
        self.formLayout_5.setObjectName("formLayout_5")
        self.sigmaMLabel = QtWidgets.QLabel(self.groupBox_6)
        self.sigmaMLabel.setObjectName("sigmaMLabel")
        self.formLayout_5.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.sigmaMLabel)
        self.sigmaMLineEdit = QtWidgets.QLineEdit(self.groupBox_6)
        self.sigmaMLineEdit.setObjectName("sigmaMLineEdit")
        self.formLayout_5.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.sigmaMLineEdit)
        self.epsilonJLabel = QtWidgets.QLabel(self.groupBox_6)
        self.epsilonJLabel.setObjectName("epsilonJLabel")
        self.formLayout_5.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.epsilonJLabel)
        self.epsilonJLineEdit = QtWidgets.QLineEdit(self.groupBox_6)
        self.epsilonJLineEdit.setObjectName("epsilonJLineEdit")
        self.formLayout_5.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.epsilonJLineEdit)
        self.reset_ia_LJ = QtWidgets.QToolButton(self.groupBox_6)
        self.reset_ia_LJ.setObjectName("reset_ia_LJ")
        self.formLayout_5.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.reset_ia_LJ)
        self.verticalLayout_4.addWidget(self.groupBox_6)
        self.groupBox_8 = QtWidgets.QGroupBox(self.otherPage)
        self.groupBox_8.setObjectName("groupBox_8")
        self.formLayout_7 = QtWidgets.QFormLayout(self.groupBox_8)
        self.formLayout_7.setObjectName("formLayout_7")
        self.r_cut_aLabel = QtWidgets.QLabel(self.groupBox_8)
        self.r_cut_aLabel.setObjectName("r_cut_aLabel")
        self.formLayout_7.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.r_cut_aLabel)
        self.r_cut_aLineEdit = QtWidgets.QLineEdit(self.groupBox_8)
        self.r_cut_aLineEdit.setObjectName("r_cut_aLineEdit")
        self.formLayout_7.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.r_cut_aLineEdit)
        self.r_cut_bLabel = QtWidgets.QLabel(self.groupBox_8)
        self.r_cut_bLabel.setObjectName("r_cut_bLabel")
        self.formLayout_7.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.r_cut_bLabel)
        self.r_cut_bLineEdit = QtWidgets.QLineEdit(self.groupBox_8)
        self.r_cut_bLineEdit.setObjectName("r_cut_bLineEdit")
        self.formLayout_7.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.r_cut_bLineEdit)
        self.r_cut_abLabel = QtWidgets.QLabel(self.groupBox_8)
        self.r_cut_abLabel.setObjectName("r_cut_abLabel")
        self.formLayout_7.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.r_cut_abLabel)
        self.r_cut_abLineEdit = QtWidgets.QLineEdit(self.groupBox_8)
        self.r_cut_abLineEdit.setObjectName("r_cut_abLineEdit")
        self.formLayout_7.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.r_cut_abLineEdit)
        self.resetRcut = QtWidgets.QToolButton(self.groupBox_8)
        self.resetRcut.setObjectName("resetRcut")
        self.formLayout_7.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.resetRcut)
        self.verticalLayout_4.addWidget(self.groupBox_8)
        CreateModel.addPage(self.otherPage)
        self.gridWidthNumberOfAtomsLabel.setBuddy(self.gridWidth)
        self.gridHeightNumberOfAtomsLabel.setBuddy(self.gridHeight)
        self.distanceBetweenAtomsLabel.setBuddy(self.distanceBetweenAtoms)
        self.xPeriodicBoundariesLabel.setBuddy(self.xPeriodicBoundariesCheckBox)
        self.yPeriodicBoudariesLabel.setBuddy(self.yPeriodicBoudariesCheckBox)
        self.firstSpeciesMoleFractionLabel.setBuddy(self.firstSpeciesMoleFraction)
        self.temperatureKLabel.setBuddy(self.temperatureKDoubleSpinBox)
        self.timestepLabel.setBuddy(self.timestepLineEdit)
        self.sigmaMLabel.setBuddy(self.sigmaMLineEdit)
        self.epsilonJLabel.setBuddy(self.epsilonJLineEdit)

        self.retranslateUi(CreateModel)
        QtCore.QMetaObject.connectSlotsByName(CreateModel)
        CreateModel.setTabOrder(self.gridWidth, self.gridHeight)
        CreateModel.setTabOrder(self.gridHeight, self.keepRatioCheckBox)
        CreateModel.setTabOrder(self.keepRatioCheckBox, self.distanceBetweenAtoms)
        CreateModel.setTabOrder(self.distanceBetweenAtoms, self.boxWidthLineEdit)
        CreateModel.setTabOrder(self.boxWidthLineEdit, self.xPeriodicBoundariesCheckBox)
        CreateModel.setTabOrder(self.xPeriodicBoundariesCheckBox, self.yPeriodicBoudariesCheckBox)
        CreateModel.setTabOrder(self.yPeriodicBoudariesCheckBox, self.firstSpeciesMoleFraction)
        CreateModel.setTabOrder(self.firstSpeciesMoleFraction, self.previewButton)
        CreateModel.setTabOrder(self.previewButton, self.temperatureKDoubleSpinBox)
        CreateModel.setTabOrder(self.temperatureKDoubleSpinBox, self.timestepLineEdit)
        CreateModel.setTabOrder(self.timestepLineEdit, self.sigmaMLineEdit)
        CreateModel.setTabOrder(self.sigmaMLineEdit, self.epsilonJLineEdit)
        CreateModel.setTabOrder(self.epsilonJLineEdit, self.reset_ia_LJ)

    def retranslateUi(self, CreateModel):
        _translate = QtCore.QCoreApplication.translate
        CreateModel.setWindowTitle(_translate("CreateModel", "Create Model"))
        self.speciesPage.setTitle(_translate("CreateModel", "Species definition"))
        self.groupBox_2.setTitle(_translate("CreateModel", "First species"))
        self.groupBox_1.setTitle(_translate("CreateModel", "Second species"))
        self.spatialPage.setTitle(_translate("CreateModel", "Spatial configuration"))
        self.groupBox.setTitle(_translate("CreateModel", "Grid"))
        self.gridWidthNumberOfAtomsLabel.setText(_translate("CreateModel", "Grid width (number of atoms)"))
        self.gridHeightNumberOfAtomsLabel.setText(_translate("CreateModel", "Grid height (number of atoms)"))
        self.keepRatioLabel.setText(_translate("CreateModel", "Keep equal ratio"))
        self.distanceBetweenAtomsLabel.setText(_translate("CreateModel", "Distance between atoms (m)"))
        self.label.setText(_translate("CreateModel", "Atom number"))
        self.label_atom_number.setText(_translate("CreateModel", "10000"))
        self.boxWidthLabel.setText(_translate("CreateModel", "Box width (m)"))
        self.groupBox_4.setTitle(_translate("CreateModel", "Boundaries"))
        self.xPeriodicBoundariesLabel.setText(_translate("CreateModel", "Enable X periodic boundaries"))
        self.yPeriodicBoudariesLabel.setText(_translate("CreateModel", "Enable Y periodic boundaries"))
        self.groupBox_3.setTitle(_translate("CreateModel", "Mixup"))
        self.firstSpeciesMoleFractionLabel.setText(_translate("CreateModel", "First species mole fraction"))
        self.previewButton.setText(_translate("CreateModel", "Preview"))
        self.otherPage.setTitle(_translate("CreateModel", "Other parameters"))
        self.groupBox_5.setTitle(_translate("CreateModel", "Thermodynamics"))
        self.temperatureKLabel.setText(_translate("CreateModel", "Temperature (K)"))
        self.groupBox_7.setTitle(_translate("CreateModel", "Simulation"))
        self.timestepLabel.setText(_translate("CreateModel", "Timestep (s)"))
        self.resetTimestep.setText(_translate("CreateModel", "Reset"))
        self.groupBox_6.setTitle(_translate("CreateModel", "Inter-atomic Lennard-Jones parameters"))
        self.sigmaMLabel.setText(_translate("CreateModel", "Sigma (m)"))
        self.epsilonJLabel.setText(_translate("CreateModel", "Epsilon (J)"))
        self.reset_ia_LJ.setText(_translate("CreateModel", "Reset"))
        self.groupBox_8.setTitle(_translate("CreateModel", "Performance"))
        self.r_cut_aLabel.setText(_translate("CreateModel", "r_cut_a (m)"))
        self.r_cut_bLabel.setText(_translate("CreateModel", "r_cut_b (m)"))
        self.r_cut_abLabel.setText(_translate("CreateModel", "r_cut_ab (m)"))
        self.resetRcut.setText(_translate("CreateModel", "Reset"))




# -*-encoding: utf-8 -*-
def fcntl(fd, op, arg=0):
    return 0

def ioctl(fd, op, arg=0, mutable_flag=True):
    if mutable_flag:
        return 0
    else:
        return ""

def flock(fd, op):
    return

def lockf(fd, operation, length=0, start=0, whence=0):
    return



# -*-encoding: utf-8 -*-
"""
Model builder.
Stores and defines physical properties of a group of atoms.

The model handles two species of atom (one of them can be ignored by setting the correct mole fraction).
The first species values are stored in the first part of arrays (lower indices), the second species in what lasts
(higher indices). This is meant to facilitate computation of inter-atomic forces and potential energy.
"""

import numexpr as ne
from ..utils.data_mng import *


class Model:
    """

    Parameters
    ----------
    pos : np.array
        Atoms' position.

        Note
        ----
        If `v` is not set, it is initialized as an array full of zeros.
    v : np.array
        Atoms' speed.

        Note
        ----
        Taken in account only if `pos` is set.
    npart : int
        Total number of atoms.

        Note
        ----
        Setting `pos` overrides this.
    x_a : float
        Mole fraction of species A.

    Attributes
    ----------

    T : float
        Temperature.
        Is calculated from the average kinetic energy of the atoms.
        May be set to any positive value, in which case atoms' speed will be scaled to match the desired temperature.
    EC : float
        Microscopic kinetic energy.

        Note
        ----
        Cannot be changed as-is, but setting :py:attr:`T` is one way to do so.
    total_EC : float
        Total kinetic energy.

        Note
        ----
        Cannot be changed as-is.
    kB : float
        Boltzmann constant. If changed, will affect the way the model behaves regarding temperature.
    pos
    v : np.array
        List of atom positions and speeds. First axis represents the atom, second axis the dimension (x or y).
    dt : float
        Timestep used for simulation.

        Note
        ----
        An acceptable value is calculated when species are defined, but it may be set to anything else.
    npart : int
        Total number of atoms.
    x_a : float
        Mole fraction of species A.
    n_a : int
        Atom number for species A, calculated from :py:attr:`x_a` and :py:attr:`npart` If set, :py:attr:`x_a` will be
        recalculated.
    epsilon_a
    epsilon_b : float
        Epsilon value (J, in Lennard-Jones potential) for species a or b.
    sigma_a
    sigma_b : float
        Sigma value (m, in Lennard-Jones potential) for species a or b.
    epsilon_ab
    sigma_ab : float
        Inter-species epsilon and sigma values.

        Note
        ----
        Cannot be changed as-is. If you want to change these values, modify the corresponding items in the
        :py:attr:`params` dictionary.
    re_a
    re_b
    re_ab : float
        Estimated radius of minimum potential energy.
    rcut_fact : float
        When the model is simulated, atoms further than :code:`rcut_fact*re` do not interact. Defaults to `2.0`.
    params : dict
        Model parameters, needed for the simulation.

        Warning
        -------
        Changing directly these values may lead to unpredicted behaviour if not documented.
    kong : dict
        Kong rules to estimate inter-species sigma and epsilon parameters.
    inter_species_rule : dict
        Rules to automatically estimate inter-species sigma and epsilon parameters. Defaults to :py:attr:`kong`.
    x_lim_inf
    y_lim_inf : float
        Lower x and y position of the box boundaries.
    x_lim_sup
    y_lim_sup : float
        Upper x and y position of the box boundaries.
    length_x
    length_y : float
        Size of the box along x and y axis. Those parameter are tied with `*_lim_***` and any change on one of them is
        correctly taken in account.
    lim_inf
    lim_sup
    length : np.array
        2 elements wide array containing corresponding :code:`(x_*, y_*)` values.

        Note
        ----
        Cannot be changed as-is.
    x_periodic
    y_periodic : int
        Defines periodic conditions for x and y axis. Set to 1 to define periodic boundary condition or 0 to live in
        an infinite empty space.
        Defaults to `0`.
    mass : float
        Total mass in the model.

        Note
        ----
        Cannot be changed as-is.
    m : np.array
        Mass of each atom. Shape is :code:`(npart, 2)` in order to facilitate calculations of kinetic energy and
        Newton's second law.

        Warning
        -------
        You should not change those values unless you know what you are doing.

    """

    kong = {
        "sigma":"(  (epsilon_a * sigma_a**12 * (1 + ((epsilon_b*sigma_b**12)/(epsilon_a*sigma_a**12))**(1/13))**13)  /   (2**13 * np.sqrt(epsilon_b*sigma_b**6*epsilon_a*sigma_a**6)) )**(1/6)",
        "epsilon":"np.sqrt(epsilon_b*sigma_b**6*epsilon_a*sigma_a**6)/sigma_ab**6",
    }

    kB = 1.38064852e-23 # en unités SI

    _special_values_f = {}
    _derived_values_f = {}

    def __init__(self, pos=None, v=None, npart=0, x_a=1.0):

        for f in self._special_values:
            self._special_values_f[f] = eval("self.set_" + f)

        for f in self._derived_values:
            self._derived_values_f[f] = eval("self.get_" + f)

        if pos:
            self.pos = pos
            npart = len(pos)
        else:
            self.pos = np.zeros((npart,2))

        if v and pos:
            self.v = v
        else:
            self.v = np.zeros((npart,2))

        self.inter_species_rule = self.kong

        self.params = {
            "npart":npart,
            "n_a":npart,
            "rcut_fact":2.0,
            "x_periodic":0,
            "y_periodic":0,
            "x_lim_inf":0,
            "y_lim_inf":0,
            "gamma":0.5,
            "up_zone_lower_limit":0.0,
            "up_apply_force_x":0,
            "up_apply_force_y":0,
            "low_zone_upper_limit":0.0,
            "low_block":0,
        }

        self.x_lim_sup = 0
        self.y_lim_sup = 0
        self.set_ab((1,1,1),(1,1,1))
        self.x_a = x_a
        self.set_dt()
        self.set_periodic_boundary()

    def copy(self):
        """

        Returns
        -------
        builder.Model
            Copy of the current model
        """
        m = Model()
        m.pos = self.pos.copy()
        m.v = self.v.copy()
        m.params = self.params.copy()
        m._m()
        return m

    __copy__ = copy

    _derived_values = [  # Les valeurs calculables à partir des autres
        "T",
        "EC",
        "total_EC",
        "mass",
        "length",
        "lim_sup",
        "lim_inf",
        "up_forces",
    ]

    def __getattr__(self, item):
        try:
            return self._derived_values_f[item]()
        except KeyError:
            try:
                return self.params[item]
            except KeyError:
                raise AttributeError(item)

    _special_values=[ # Les valeurs à vérifier ou  à transformer avant enregistrement
        "T",
        "x_a",
        "n_a",
        "x_periodic",
        "y_periodic",
        "x_lim_sup",
        "x_lim_inf",
        "y_lim_sup",
        "y_lim_inf",
        "length_x",
        "length_y",
        "dt",
    ]

    def __setattr__(self, key, value):
        try:
            self._special_values_f[key](value)
        except KeyError:
            super(Model, self).__setattr__(key, value)

    def _set_species(self, epsilon : float, sigma : float, m : float, sp : str):
        self.params["epsilon_"+sp] = epsilon
        self.params["sigma_"+sp] = sigma
        self.params["re_"+sp] = 2.0**(1.0/6.0)*sigma
        self.params["rcut_"+sp] = self.rcut_fact*self.params["re_"+sp]
        self.params["m_"+sp] = m

    def set_a(self, epsilon : float, sigma : float, m : float):
        """
        Sets species a parameters.

        Parameters
        ----------
        epsilon : float
            Epsilon in Lennard-Jones potential.
        sigma : float
            Sigma in Lennard-Jones potential.
        m : float
            Mass of the atom.

        Returns
        -------

        """
        self._set_species(epsilon, sigma, m, "a")

    def set_b(self, epsilon : float, sigma : float, m : float):
        """
        Same as :py:attr:`set_a`.

        Returns
        -------

        """
        self._set_species(epsilon, sigma, m, "b")

    def calc_ab(self):
        epsilon_a = self.epsilon_a
        epsilon_b = self.epsilon_b
        sigma_a = self.sigma_a
        sigma_b = self.sigma_b

        rule = self.inter_species_rule
        sigma_ab = eval(rule["sigma"]) # on en a besoin pour le calcul d'epsilon_ab
        epsilon_ab = eval(rule["epsilon"])
        self._set_species(epsilon_ab, sigma_ab, 0, "ab")

        self.params["re"] = max(self.re_a, self.re_b)

        self._m()

    def set_ab(self, a : tuple, b : tuple):
        """
        Sets species a and b parameters, and calculates inter-species parameters.

        Parameters
        ----------
        a : tuple
            First species parameters, under the form :code:`(epsilon, sigma, mass)`
        b : tuple
            Second species parameters, under the same form.
        Returns
        -------

        """
        self.set_a(*a)
        self.set_b(*b)

        self.calc_ab()

    def set_n_a(self,n_a : int):
        self.x_a = n_a/self.npart

    def atom_grid(self, n_x : int, n_y : int, d : float):
        """
        Creates a grid containing :code:`n_x*n_y` atoms.

        Sets :py:attr:`npart`, :py:attr:`pos`, :py:attr:`dt`, :py:attr:`v` and :py:attr:`m`.

        Parameters
        ----------
        n_x : int
            number of columns
        n_y : int
            number of rows
        d : float
            inter-atomic distance

        Returns
        -------

        """
        self.params["npart"] = n_x*n_y
        self.params["n_a"] = int(self.x_a*self.npart)

        posx = np.concatenate([np.arange(0,n_x,1.0) for i in range(n_y)])*d
        posy = np.concatenate([i*np.ones(n_x) for i in range(n_y)])*d
        self.pos = np.transpose([posx, posy])

        self.v = np.zeros(self.pos.shape)

        self._m()

        self.x_lim_sup = (n_x - 0.5)*d
        self.y_lim_sup = (n_y - 0.5)*d
        self.x_lim_inf = -0.5*d
        self.y_lim_inf = -0.5*d

        self.set_dt()

    def _ord_lim_x(self): # s'assure que la taille de la boîte est positive
        self.params["x_lim_inf"], self.params["x_lim_sup"] = sorted((self.x_lim_inf, self.x_lim_sup))
        self.params["length_x"] = self.x_lim_sup - self.x_lim_inf

    def _ord_lim_y(self):
        self.params["y_lim_inf"], self.params["y_lim_sup"] = sorted((self.y_lim_inf, self.y_lim_sup))
        self.params["length_y"] = self.y_lim_sup - self.y_lim_inf

    def set_x_lim_inf(self,x_lim_inf : float):
        self.params["x_lim_inf"] = x_lim_inf
        self._ord_lim_x()

    def set_y_lim_inf(self,y_lim_inf : float):
        self.params["y_lim_inf"] = y_lim_inf
        self._ord_lim_y()

    def get_lim_inf(self):
        return np.array([self.x_lim_inf, self.y_lim_inf])

    def get_lim_sup(self):
        return np.array([self.x_lim_sup, self.y_lim_sup])

    def set_x_lim_sup(self,x_lim_sup : float):
        self.params["x_lim_sup"] = x_lim_sup
        self._ord_lim_x()

    def set_y_lim_sup(self,y_lim_sup : float):
        self.params["y_lim_sup"] = y_lim_sup
        self._ord_lim_y()

    def set_length_x(self,length_x : float):
        self.x_lim_sup = self.x_lim_inf + length_x

    def set_length_y(self,length_y : float):
        self.y_lim_sup = self.y_lim_inf + length_y

    def get_length(self):
        return np.array([self.length_x, self.length_y])

    def shuffle_atoms(self): # mélange les atomes aléatoirement pour répartir les 2 espèces dans l'espace
        """
        Shuffle atoms' position in order to easily create a homogeneous repartition of the two species.
        Should be called just right after the positions are defined.

        Note
        ----
        Atoms' speed is not shuffled.

        Returns
        -------

        """
        np.random.shuffle(self.pos)

    def random_speed(self): # donne une vitesse aléatoire aux atomes pour une température non nulle
        """
        Gives a random speed to the atoms, following a normal law, in order to have a strictly positive temperature.

        Returns
        -------

        """
        self.v = np.random.normal(size=(self.npart,2))

    def _m(self): # vecteur de masses
        """
        Constructs :py:attr:`m`.

        Returns
        -------
        m : np.array

        """
        m = np.concatenate((self.m_a*np.ones(self.n_a), self.m_b*np.ones(self.npart - self.n_a)))
        self.m = np.transpose([m,m])
        return self.m

    def get_up_forces(self):
        return np.array([self.up_x_component, self.up_y_component])

    def get_total_EC(self): # énergie cinétique totale
        v = self.v
        m = self.m
        return 0.5*ne.evaluate("sum(m*v**2)")

    def get_EC(self): # énergie cinétique microscopique
        v = self.v
        m = self.m
        v_avg = np.average(v, axis=0)
        return 0.5*ne.evaluate("sum(m*(v-v_avg)**2)")

    def get_T(self):
        v = self.v
        m = self.m
        v_avg = np.average(v, axis=0)
        return self.EC/(self.kB*self.npart)

    def set_T(self, T : float):
        T = max(0,T)
        if not self.T:
            self.random_speed()
        v = self.v
        v_avg = np.average(v, axis=0)
        self.v = v_avg+v*np.sqrt(T/self.T)

    def get_mass(self):
        return self.n_a*self.m_a + (self.npart-self.n_a)*self.m_b

    def set_x_a(self,x_a : float):
        self.params["x_a"] = min(max(x_a,0.0),1.0)
        self.params["n_a"] = int(x_a*self.npart)
        self._m()

    def set_x_periodic(self, x : int = 1):
        if x: # pour être certains de la valeur de x
            x = 1
        else:
            x = 0
        self.params["x_periodic"] = x

    def set_y_periodic(self, y : int = 1):
        if y:
            y = 1
        else:
            y = 0
        self.params["y_periodic"] = y

    def set_periodic_boundary(self, x : int = 1, y : int = 1): # conditions périodiques de bord : 1, sinon 0
        """
        Set periodic boundaries on both axis.

        Parameters
        ----------
        x : int
            see :py:attr:`x_periodic`
        y : int
            see :py:attr:`y_periodic`

        Returns
        -------

        """
        self.x_periodic = x
        self.y_periodic = y

    def decent_dt(self):
        epsilon = max(self.epsilon_a, self.epsilon_b, self.epsilon_ab)
        sigma = min(self.sigma_a, self.sigma_b, self.sigma_ab)
        m = min(self.m_a, self.m_b)
        freq = np.sqrt((57.1464 * epsilon / (sigma**2)) / m) / (2*np.pi)
        period = 1 / freq
        dt = period / 50
        return dt

    def set_dt(self, dt : int = None):
        """
        Defines the timestep used for simulation.

        Parameters
        ----------
        dt : float
            Desired timestep. If not set, will be calculated from species a's properties.
        Returns
        -------

        """
        if not dt: # période d'oscillation pour pouvoir calibrer le pas de temps
            dt = self.decent_dt()

        self.params["dt"] = abs(dt)





# -*-encoding: utf-8 -*-
"""
Simulator.
Simulates the dynamics of a model (on the CPU or GPU).
"""

import numpy as np
import numexpr as ne
import scipy.interpolate as inter
import warnings

from .forces_CPU import ForcesComputeCPU
from .forces_GPU import ForcesComputeGPU

class Simulation:
    """
    Simulator for a model.

    Parameters
    ----------
    model : builder.Model
        Model to simulate.
        The original model object is copied, and thus preserved, which allows it to serve as a reference.
    simulation : Simulation
        Simulation to copy.
        The simulation is copied, and the computing module reinitialized.
    prefer_gpu : bool
        Specifies if GPU should be used to compute inter-atomic forces.
        Defaults to `True`, as it generally results in a significant speed gain.

    Attributes
    ----------
    model : builder.Model
        Model that is simulated.

        -------
        Warning
        -------
        Works as usual, but changing its properties will not affect the simulation as intended.
        You should construct another simulation from this model to correctly take in account the changes.
        This is due to the fact that some values (eg. Lennard-Jones parameters) are treated as constants during shader
        initialisation, in order to speed up the calculations.
    current_iter : int
        Number of iterations already computed, since initialisation.
    context : moderngl.Context
        ModernGL context used to build and run compute shader.
    F : numpy.ndarray
        Last computed forces applied to atoms. Initialized to zeros.

        Warning
        -------
        Changing the values will affect behavior of the model.
    """

    def __init__(self, model = None, simulation = None, prefer_gpu = True):

        if simulation:
            model = simulation.model

        self.model = model.copy()

        # paramétrage du module de calcul
        consts = dict()
        for k in model.params:
            consts[k.upper()] = model.params[k]

        if prefer_gpu:
            try:
                self._compute = ForcesComputeGPU(consts)
            except:
                warnings.warn("GPU not available, falling back on CPU. GPU compute needs OpenGL >=4.3.")
                self._compute = ForcesComputeCPU(consts)
        else:
            self._compute = ForcesComputeCPU(consts)

        self.T_f = lambda t:self.T[-1]
        self.Fx_f = lambda t:0.0
        self.Fy_f = lambda t:0.0

        if simulation :
            self.current_iter = simulation.current_iter

            self.state_fct = simulation.state_fct

            self.T_cntl = simulation.T_cntl
            if self.T_cntl:
                self.T_f = simulation.T_f
            self.Fx_f = simulation.Fx_f
            self.Fy_f = simulation.Fy_f

            self.F = simulation.F
        else:
            self.current_iter = 0

            self.state_fct = dict()

            self.state_fct["T"] = []
            self.state_fct["T_ctrl"] = []
            self.state_fct["T_ramps"] = [[],[]]

            self.state_fct["Fx_ramps"] = [[],[]]
            self.state_fct["Fy_ramps"] = [[],[]]

            self.state_fct["EC"] = []
            self.state_fct["EP"] = []
            self.state_fct["ET"] = []
            self.state_fct["bonds"] = []

            self.state_fct["time"] = []
            self.state_fct["iters"] = []

            self.T_cntl = False

            self.F = np.zeros(self.model.pos.shape) # Doit être initialisé et conservé d'une itération à l'autre

        for s in self.state_fct:
            self.__setattr__(s, self.state_fct[s])

    def iter(self, n=1, callback=None):
        """
        Iterates one or more simulation steps.

        Basically, follows the Position-Verlet method to update positions and speeds, and computes inter-atomic forces
        derived from Lennard-Jones potential.

        Parameters
        ----------
        n: int
            Number of iterations to perform.
        callback : callable
            A callback function that must take the Simulation object as first argument.
            It is called at the end of each iteration.

        Note
        ----
        Setting n is significantly faster than calling :py:meth:`iter` several times.

        Example
        -------
        .. code-block:: python

            model.iter(5)

        Returns
        -------

        """

        for s in self.state_fct:
            self.__setattr__(s, self.state_fct[s])

        betaC = self.T_cntl # Contrôle de la température

        # on crée des alias aux valeurs du modèles pour numexpr
        v = self.model.v
        pos = self.model.pos
        dt = self.model.dt
        dt2 = dt/2.0
        m = self.model.m
        dtm = dt/m
        npart = self.model.npart
        inv2npart = 0.5/npart
        knparts = self.model.kB * npart
        gamma = self.model.gamma

        limInf = self.model.lim_inf
        limSup = self.model.lim_sup
        length = self.model.length

        F = self.F
        bondsGL = np.zeros(self.model.npart)

        periodic = self.model.x_periodic or self.model.y_periodic

        apply_up_zone_forces = self.model.up_apply_force_x or self.model.up_apply_force_y
        #up_zone_force = self.model.up_forces
        up_zone_limit = self.model.up_zone_lower_limit
        low_zone_block = self.model.low_block
        low_zone_limit = self.model.low_zone_upper_limit

        if periodic:
            length *= (self.model.x_periodic, self.model.y_periodic)
            # on n'applique les conditions périodiques que selon le(s) axe(s) spécifié(s)

        kick = "F"
        micro_ke = "sum(m*(v-v_avg)**2)"
        if apply_up_zone_forces:
            kick = "(F+up_mask*up_zone_force)"
            up_mask = np.zeros(pos.shape)

        compute_rotative_term = apply_up_zone_forces and not(self.model.y_periodic)
        if compute_rotative_term:
            micro_ke = "sum(m*(v-v_avg-rotative_term)**2)"
            rotative_term = np.zeros(pos.shape)
            y_middle = (self.model.y_lim_sup + self.model.y_lim_inf)/2
            from_y_middle = np.zeros(npart)
        kick = "(v + ("+kick+"*dtm))"
        if betaC:
            kick += "*sqrt(1 + gamma*(T_v/T - 1))"
        if low_zone_block:
            # On présélectionne les atomes bloqués, afin que leur nombre ne change pas
            low_block_mask = pos[:,1] > low_zone_limit
            low_block_mask = np.array([low_block_mask]*2).T
            kick += "*low_block_mask"

        for i in range(n):

            t = self.current_iter * dt # l'heure, qui sert pour le calcul de température

            ne.evaluate("pos + v*dt2", out=pos)  # half drift

            # conditions périodiques de bord
            if periodic:
                ne.evaluate("pos + (pos<limInf)*length - (pos>limSup)*length", out=pos)

            self._compute.set_pos(pos)

            v_avg = np.average(v, axis=0)

            if apply_up_zone_forces: # recalcul du masque, parce que ça bouge
                up_zone_force = self.F_f(t)
                up_mask[:,:] = np.array([pos[:,1] > up_zone_limit]*2).T
            if compute_rotative_term:
                from_y_middle[:] = pos[:,1]-y_middle
                rotative_term[:,0] = (np.sum(v[:,0]/from_y_middle)/npart)*from_y_middle

            # Énergie cinétique et température
            EC = 0.5 * ne.evaluate(micro_ke)
            T = EC / knparts
            self.EC.append(EC)
            self.T.append(T)

            F[:] = self._compute.get_F()

            # Énergie potentielle
            EPgl = self._compute.get_PE()
            EP = 0.5 * ne.evaluate("sum(EPgl)")
            self.EP.append(EP)
            self.ET.append(EC + EP)

            # Thermostat
            T_v = self.T_f(t)
            self.T_ctrl.append(T_v)
            ne.evaluate(kick, out=v) # kick

            ne.evaluate("pos + v*dt2", out=pos)  # half drift

            bondsGL[:] = self._compute.get_COUNT()
            self.bonds.append(inv2npart*ne.evaluate("sum(bondsGL)"))

            self.iters.append(self.current_iter)
            self.time.append(t)

            if callback:
                callback(self)

            self.current_iter += 1

    def _f(self, t, y):
        f2 = inter.interp1d(t, y)

        def f(x):
            if x<t[0]:
                return y[0]
            elif x>t[-1]:
                return y[-1]
            else:
                return float(f2(x)) # pour la consistance des types

        return f

    def set_T_f(self, f):
        """
        Sets function that controls temperature.

        Parameters
        ----------
        f : callable
            Must take time (float) as an argument and return temperature (in K, float).

        Returns
        -------

        """
        self.T_cntl = True
        self.T_f = f
        self.model.T = f(self.current_iter*self.model.dt)

    def set_T_ramps(self, t, T):
        """
        Creates a function based on ramps and uses it for temperature control.
        Values of the function are interpolated between points given in `t` and `T`.
        Temperature is supposed constant before the first point and after the last one.

        Parameters
        ----------
        t : array
            Time.
        T : array
            Associated temperatures.

        Returns
        -------

        """
        if len(t)>1:
            self.state_fct["T_ramps"] = [list(t), list(T)]
            self.T_ramps = self.state_fct["T_ramps"]
            self.set_T_f(self._f(t, T))

    def F_f(self, t):
        """

        Parameters
        ----------
        t
            Time in seconds.
        Returns
        -------
        numpy.ndarray
              External forces applied at time `t` (2-component vector).
        """
        return np.array((self.Fx_f(t), self.Fy_f(t)))

    def set_Fx_ramps(self, t, Fx):
        """
        Creates a function based on ramps and uses it for external forces control along axis x.
        See `set_T_ramps` for further details.

        Parameters
        ----------
        t : array
            Time.
        Fx : array
            Associated forces.
        """
        if len(t)>1:
            self.state_fct["Fx_ramps"] = [list(t), list(Fx)]
            self.Fx_ramps = self.state_fct["Fx_ramps"]
            self.Fx_f = self._f(t, Fx)

    def set_Fy_ramps(self, t, Fy):
        """
        Creates a function based on ramps and uses it for external forces control along axis y.
        See `set_T_ramps` for further details.

        Parameters
        ----------
        t : array
            Time.
        Fy : array
            Associated forces.
        """
        if len(t)>1:
            self.state_fct["Fy_ramps"] = [list(t), list(Fy)]
            self.Fy_ramps = self.state_fct["Fy_ramps"]
            self.Fy_f = self._f(t, Fy)


from PyQt5.QtWidgets import QWizard, QMessageBox
from .qt.create_model import Ui_CreateModel
from .species_params import species_params

from ..simulation.builder import Model
from .model_viewer import ModelView

from ._conv import _float

import matplotlib
matplotlib.use('Qt5Agg')

class CreateModelDialog(QWizard):

    def __init__(self, parent_window=None):
        super().__init__()
        self.ui = Ui_CreateModel()
        self.ui.setupUi(self)

        self.parent_window = parent_window

        # Panneau espèces
        self.species_a_params = species_params()
        self.species_b_params = species_params()
        self.ui.layout_a.addWidget(self.species_a_params)
        self.ui.layout_b.addWidget(self.species_b_params)

        self.species_a_params.editingFinished.connect(self.check_species)
        self.species_b_params.editingFinished.connect(self.check_species)

        self.model = Model()

        # Panneau conf spatiale
        self.ui.speciesPage.validatePage =  self.to_spatial_conf

        self.ui.keepRatioCheckBox.stateChanged.connect(self.keep_ratio)

        self.ui.gridWidth.valueChanged.connect(self.grid_w_changed)
        self.ui.gridHeight.valueChanged.connect(self.grid_h_changed)

        self.ui.distanceBetweenAtoms.editingFinished.connect(self.checked_distance)
        self.ui.boxWidthLineEdit.editingFinished.connect(self.box_w_changed)

        self.ui.xPeriodicBoundariesCheckBox.stateChanged.connect(self.model.set_x_periodic)
        self.ui.yPeriodicBoudariesCheckBox.stateChanged.connect(self.model.set_y_periodic)

        self.ui.firstSpeciesMoleFraction.valueChanged.connect(self.model.set_x_a)

        self.ui.previewButton.clicked.connect(self.preview)

        # Panneau paramètres divers
        self.ui.spatialPage.validatePage = self.set_parameters

        self.ui.reset_ia_LJ.clicked.connect(self.reset_es_ab)

        self.ui.resetTimestep.clicked.connect(self.set_decent_timestep)

        self.ui.temperatureKDoubleSpinBox.valueChanged.connect(self.model.set_T)

        self.ui.timestepLineEdit.editingFinished.connect(self.checked_timestep)
        self.ui.epsilonJLineEdit.editingFinished.connect(self.check_es_ab)
        self.ui.sigmaMLineEdit.editingFinished.connect(self.check_es_ab)

        self.ui.r_cut_aLineEdit.editingFinished.connect(self.check_rcut)
        self.ui.r_cut_bLineEdit.editingFinished.connect(self.check_rcut)
        self.ui.r_cut_abLineEdit.editingFinished.connect(self.check_rcut)

        self.ui.resetRcut.clicked.connect(self.reset_rcut)

        self.show()

    # Panneau espèces

    def check_species(self):
        t = (self.species_a_params, self.species_b_params)
        a, b = (s.check_values() for s in t)
        if a ^ b:
            t[a].set_values(t[b].get_values(), False)
        return a or b

    # Panneau conf spatiale

    def to_spatial_conf(self):
        if not self.check_species():
            qmb = QMessageBox()
            qmb.setText("Please define a species.")
            qmb.exec()
            return False

        self.model.params["display_name"] = " ".join([
            self.species_a_params.ui.nameLineEdit.text(),
            "and",
            self.species_b_params.ui.nameLineEdit.text(),
        ])
        self.model.set_ab(
            self.species_a_params.get_values(),
            self.species_b_params.get_values()
        )
        self.checked_distance()
        return True

    def keep_ratio(self, b):
        b = bool(b)
        self.ui.gridHeight.setReadOnly(b)
        self.ui.gridHeight.setValue(self.ui.gridWidth.value())

    def grid_w_changed(self, v):
        if self.ui.gridHeight.isReadOnly():
            self.ui.gridHeight.setValue(v)
        self.ui.label_atom_number.setText(str(v * self.ui.gridHeight.value()))
        self.checked_distance()

    def grid_h_changed(self, v):
        self.ui.label_atom_number.setText(str(v * self.ui.gridWidth.value()))

    def box_w_changed(self):
        try:
            l = _float(self.ui.boxWidthLineEdit)
        except:
            l = _float(self.ui.distanceBetweenAtoms) * (self.ui.gridWidth.value())
        self.ui.boxWidthLineEdit.setText(str(l))
        self.ui.distanceBetweenAtoms.setText(str(l / (self.ui.gridWidth.value())))

    def checked_distance(self):
        d = _float(self.ui.distanceBetweenAtoms, self.model.re_ab)
        self.ui.distanceBetweenAtoms.setText(str(d))

        self.ui.boxWidthLineEdit.setText(str(d * (self.ui.gridWidth.value())))

        return d

    def preview(self, b):
        self.model.atom_grid(self.ui.gridWidth.value(), self.ui.gridHeight.value(), self.checked_distance())
        self.model.shuffle_atoms()

        self.mv = ModelView(self.model)
        self.mv.show()

    # Panneau paramètres divers

    def reset_es_ab(self):
        self.model.calc_ab()
        self.ui.epsilonJLineEdit.setText(str(self.model.epsilon_ab))
        self.ui.sigmaMLineEdit.setText(str(self.model.sigma_ab))
        self.checked_timestep()

    def check_es_ab(self):
        epsilon_ab = _float(self.ui.epsilonJLineEdit, self.model.epsilon_ab)
        sigma_ab = _float(self.ui.sigmaMLineEdit, self.model.sigma_ab)
        self.ui.epsilonJLineEdit.setText(str(epsilon_ab))
        self.ui.sigmaMLineEdit.setText(str(sigma_ab))
        self.model.params["epsilon_ab"] = epsilon_ab
        self.model.params["sigma_ab"] = sigma_ab
        self.checked_timestep()

    def reset_rcut(self):
        rcut_a = self.model.re_a * self.model.rcut_fact
        self.ui.r_cut_aLineEdit.setText(str(rcut_a))
        rcut_b = self.model.re_b * self.model.rcut_fact
        self.ui.r_cut_bLineEdit.setText(str(rcut_b))
        rcut_ab = self.model.re_ab * self.model.rcut_fact
        self.ui.r_cut_abLineEdit.setText(str(rcut_ab))

    def check_rcut(self):
        rcut_a = _float(self.ui.r_cut_aLineEdit, self.model.rcut_a)
        self.model.params["rcut_a"] = rcut_a
        self.ui.r_cut_aLineEdit.setText(str(rcut_a))

        rcut_b = _float(self.ui.r_cut_bLineEdit, self.model.rcut_ab)
        self.model.params["rcut_b"] = rcut_b
        self.ui.r_cut_bLineEdit.setText(str(rcut_b))

        rcut_ab = _float(self.ui.r_cut_abLineEdit, self.model.rcut_ab)
        self.model.params["rcut_ab"] = rcut_ab
        self.ui.r_cut_abLineEdit.setText(str(rcut_ab))

    def set_parameters(self):
        self.model.atom_grid(self.ui.gridWidth.value(), self.ui.gridHeight.value(), self.checked_distance())
        self.model.shuffle_atoms()
        self.model.set_dt()

        self.model.T = 1.0
        self.ui.temperatureKDoubleSpinBox.setValue(1.0)

        self.reset_es_ab()

        self.ui.timestepLineEdit.setText(str(self.model.dt))

        self.reset_rcut()

        return True

    def set_decent_timestep(self):
        self.ui.timestepLineEdit.setText(str(self.model.decent_dt()))
        self.checked_timestep()

    def checked_timestep(self):
        dt = _float(self.ui.timestepLineEdit, self.model.dt)
        self.ui.timestepLineEdit.setText(str(dt))
        self.model.dt = dt

        decent = self.model.decent_dt()
        if dt > decent:
            qmb = QMessageBox()
            qmb.setText("The time-step may be too high. Simulation may then not be accurate enough.")
            qmb.exec()

        return dt

    def accept(self):
        if self.parent_window:
            self.parent_window.set_model(self.model)
        super().accept()


from PyQt5.QtWidgets import QWidget, QMessageBox

from .qt.define_external_forces import Ui_ExternalForces
from multiprocessing import Process, Queue
from . import draggableLine

from ._conv import _float


class define_exernal_forces(QWidget):

    def __init__(self, *args, simulation=None, until=0):
        super().__init__(*args)
        self.ui = Ui_ExternalForces()
        self.ui.setupUi(self)

        self.up_x_component = [[],[]]
        self.up_y_component = [[],[]]

        self.until = until

        if simulation:
            self.set_simulation(simulation)

        #self.ui.applyExternalForceCheckBox.stateChanged.connect(self.ui.up_components.setEnabled)
        self.ui.applyXUp.stateChanged.connect(self.ui.xUpBtn.setEnabled)
        self.ui.applyYUp.stateChanged.connect(self.ui.yUpBtn.setEnabled)

        self.ui.xUpBtn.clicked.connect(self.design_x_component)
        self.ui.yUpBtn.clicked.connect(self.design_y_component)

        self.ui.upFractionOfSpaceDoubleSpinBox.valueChanged.connect(self.up_spinbox_changed)
        self.ui.lowFractionOfSpaceDoubleSpinBox.valueChanged.connect(self.low_spinbox_changed)

        self.ui.lowerLimitMLineEdit.editingFinished.connect(self.up_edited)
        self.ui.upperLimitMLineEdit.editingFinished.connect(self.low_edited)

        self.show()

    def set_simulation(self, simulation):
        self.simulation = simulation
        self.model = simulation.model

        self.up_apply_x = self.model.up_apply_force_x
        self.up_apply_y = self.model.up_apply_force_y
        self.low_block = self.model.low_block

        self.up_lower_limit = self.model.up_zone_lower_limit
        self.low_upper_limit = self.model.low_zone_upper_limit

        self.up_x_component = self.simulation.state_fct["Fx_ramps"]
        self.up_y_component = self.simulation.state_fct["Fy_ramps"]

        self.up_edited()
        self.low_edited()

    def design_force(self, axis_name, default):
        queue = Queue(1)
        decent_force = self.model.epsilon_ab / self.model.sigma_ab
        axis = (0, (self.until+self.simulation.current_iter)*self.model.dt, -decent_force, decent_force)
        queue.put((axis, *default, "Time (s)", "Force, "+axis_name+" component (N)", None))
        design_thread = Process(target=draggableLine.main, args=(queue,))
        design_thread.start()
        design_thread.join()
        x_data, y_data = queue.get()
        queue.close()
        if len(x_data)<2:
            qmb = QMessageBox()
            qmb.setText("This only works with at least two points.")
            qmb.exec()
            return default
        return [x_data, y_data]

    def design_x_component(self):
        self.up_x_component = self.design_force("x", self.up_x_component)

    def design_y_component(self):
        self.up_y_component = self.design_force("y", self.up_y_component)

    def up_spinbox_changed(self):
        self.ui.lowerLimitMLineEdit.setText(str(self.c_up_lower_limit))
        self.set_up()

    def low_spinbox_changed(self):
        self.ui.upperLimitMLineEdit.setText(str(self.c_low_upper_limit))
        self.set_low()

    def up_edited(self):
        self.ui.upFractionOfSpaceDoubleSpinBox.setValue(1 - self.up_lower_limit / self.model.length_y)
        self.set_up()

    def low_edited(self):
        self.ui.lowFractionOfSpaceDoubleSpinBox.setValue(self.low_upper_limit / self.model.length_y)
        self.set_low()

    def set_up(self):
        self.model.params["up_zone_lower_limit"] = self.up_lower_limit

    def set_low(self):
        self.model.params["low_zone_upper_limit"] = self.low_upper_limit

    @property
    def c_up_lower_limit(self):
        return self.model.length_y*(1-self.ui.upFractionOfSpaceDoubleSpinBox.value())

    @property
    def c_low_upper_limit(self):
        return self.model.length_y*self.ui.lowFractionOfSpaceDoubleSpinBox.value()

    @property
    def up_lower_limit(self):
        return _float(self.ui.lowerLimitMLineEdit, self.c_up_lower_limit)

    @up_lower_limit.setter
    def up_lower_limit(self, v):
        self.ui.lowerLimitMLineEdit.setText(str(v))

    @property
    def up_apply_x(self):
        return int(self.ui.applyXUp.isChecked())

    @up_apply_x.setter
    def up_apply_x(self, b):
        self.ui.applyXUp.setChecked(b)
        self.ui.xUpBtn.setEnabled(b)

    @property
    def up_apply_y(self):
        return int(self.ui.applyYUp.isChecked())

    @up_apply_y.setter
    def up_apply_y(self, b):
        self.ui.applyYUp.setChecked(b)
        self.ui.yUpBtn.setEnabled(b)

    @property
    def low_upper_limit(self):
        return _float(self.ui.upperLimitMLineEdit, self.c_low_upper_limit)

    @low_upper_limit.setter
    def low_upper_limit(self, v):
        self.ui.upperLimitMLineEdit.setText(str(v))

    @property
    def low_block(self):
        return int(self.ui.blockCheckBox.isChecked())

    @low_block.setter
    def low_block(self, b):
        self.ui.blockCheckBox.setChecked(b)



import os
import shutil

from PyQt5.QtWidgets import QMainWindow, QTreeWidgetItem, QHeaderView, QListWidgetItem, QMessageBox, QFileDialog
from PyQt5.QtCore import QThread, pyqtSignal
from mpl_toolkits.axisartist.parasite_axes import HostAxes, ParasiteAxes
from pyqtgraph import PlotWidget
import time
import csv
import multiprocessing as mp
from datetime import timedelta

from moldyn.utils.data_mng import DynState, tmp1_path, tmp_path

MODEL_FILE_FILTER = "Model file (*.mdl);;Legacy Model file (*.zip)"

SIMULATION_FILE_FILTER_ = "Simulation file (*.mds);;\
                                                                                Legacy Simulation file (*.zip)"

try:
    mp.set_start_method('spawn')
except RuntimeError:
    pass

from multiprocessing import Process, Queue

from collections import deque

from .qt.mainwindow import Ui_MainWindow

from .create_model import CreateModelDialog
from .edit_model import EditModelDialog
from .model_viewer import ModelView

from ._conv import _float

from ..simulation.builder import Model
from ..simulation.runner import Simulation

from ..processing import visualisation as visu
from ..processing.data_proc import PDF
from . import draggableLine


class MoldynMainWindow(QMainWindow):
    updated_signal = pyqtSignal(int, float)
    movie_progress_signal = pyqtSignal(int)
    displayed_properties = dict()

    def __init__(self):
        super().__init__()

        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        # Panneau modèle
        self.ui.newModelBtn.clicked.connect(self.create_model)
        self.ui.loadModelBtn.clicked.connect(self.load_model)
        self.ui.saveModelBtn.clicked.connect(self.save_model)
        self.ui.loadSimuBtn.clicked.connect(self.load_simulation)
        self.ui.gotoSimuBtn.clicked.connect(self.goto_simu)
        self.ui.newSimuBtn.clicked.connect(self.reset_model)
        self.ui.editModelBtn.clicked.connect(self.edit_model)

        self.displayed_properties_list = {
            "display_name" : ["Name"],
            "T" : ["Temperature", "K"],
            "Lennard-Jones and mass" : {
                "First species" : {
                    "sigma_a" : ["Sigma", "m"],
                    "epsilon_a" : ["Epsilon", "J"],
                    "m_a" : ["Atomic mass", "kg"],
                    "rcut_a" : ["rcut_a", "m"],
                },
                "Second species" : {
                    "sigma_b" : ["Sigma", "m"],
                    "epsilon_b" : ["Epsilon", "J"],
                    "m_b" : ["Atomic mass", "kg"],
                    "rcut_b" : ["rcut_b", "m"],
                },
                "Inter-species" : {
                    "sigma_ab" : ["Sigma", "m"],
                    "epsilon_ab" : ["Epsilon", "J"],
                    "rcut_ab" : ["rcut_ab", "m"],
                },
            },
            "npart" : ["Atom number"],
            "x_a" : ["First species mole fraction"],
            "dt" : ["Timestep", "s"],
            "Spatial configuration" : {
                "length_x" : ["Width", "m"],
                "length_y" : ["Height", "m"],
                "x_periodic" : ["Periodic condition on x"],
                "y_periodic" : ["Periodic condition on y"],
            },
        }

        def sub_items(item, parent):
            for k in item:
                if type(item[k]) == list:
                    if len(item[k])>1:
                        item[k][1:] = ["", item[k][1]]
                    current_item = QTreeWidgetItem(item[k])
                    self.displayed_properties[k] = current_item
                else:
                    current_item = QTreeWidgetItem([k])
                    sub_items(item[k], current_item)
                parent.addChild(current_item)
                current_item.setExpanded(True)

        self.ui.paramsTreeWidget.addChild = self.ui.paramsTreeWidget.addTopLevelItem
        self.ui.paramsTreeWidget.header().setResizeMode(QHeaderView.ResizeToContents)

        sub_items(self.displayed_properties_list, self.ui.paramsTreeWidget)

        # Panneau simu

        self.ui.iterationsSpinBox.valueChanged.connect(self.update_simu_time)
        self.ui.simulationTimeLineEdit.editingFinished.connect(self.update_iters)

        self.ui.simuBtn.clicked.connect(self.simulate)

        self.updated_signal.connect(self.update_progress)

        self.ui.designTemperatureBtn.clicked.connect(self.design_temperature_profile)

        self.progress_plt = PlotWidget(self.ui.progress_groupBox)
        self.ui.progress_groupBox.layout().addWidget(self.progress_plt, 2, 0, 1, 3)
        self.progress_plt.setXRange(0,1)
        self.progress_plt.setLabel('bottom',text='Iteration')
        self.progress_plt.setLabel('left',text='Speed', units='Iteration/s')
        self.progress_gr = self.progress_plt.plot(pen='y')

        self.t_deque = deque()

        self.ui.gotoProcessBtn.clicked.connect(self.goto_process)

        self.ui.RTViewBtn.clicked.connect(self.show_model)

        # Panneau processing

        self.ui.saveRModelBtn.clicked.connect(self.save_final_model)
        self.ui.saveSimuBtn.clicked.connect(self.save_simu_history)
        self.ui.reuseModelBtn.clicked.connect(self.reuse_model)

        self.ui.exportBtn.clicked.connect(self.export_to_csv)

        self.ui.PDFButton.clicked.connect(self.PDF)

        self.temporal_variables = {
            "Time":["time","s"],
            "Temperature":["T","K"],
            "Temperature control":["T_ctrl","K"],
            "Microscopic kinetic energy":["EC","J"],
            "Inter-atomic potential energy":["EP","J"],
            "Total energy":["ET","J"],
            "Number of bonds per atom":["bonds"],
            "Iteration":["iters"],
        }

        self.temp_variables_w = []

        for v in self.temporal_variables:
            self.ui.lineComboW.addItem(v)
            i = QListWidgetItem(v)
            i.setCheckState(False)
            self.temp_variables_w.append(i)
            self.ui.lineListW.addItem(i)

        self._1d_options = {
            "log x":None,
            "log y":None,
            "grid":None,
        }
        for i_n in self._1d_options:
            i = QListWidgetItem(i_n)
            i.setCheckState(False)
            self.ui.list1DOptions.addItem(i)
            self._1d_options[i_n] = i

        self.ui.plotB.clicked.connect(self.line_graph)

        self._2d_options = {
            "density map": None,
            "deformation (compression)":None,
            "deformation (shear)":None,
            "particles": None,
        }
        for i_n in self._2d_options:
            i = QListWidgetItem(i_n)
            i.setCheckState(False)
            self.ui.surfListW.addItem(i)
            self._2d_options[i_n] = i

        self.ui.drawSurfButton.clicked.connect(self.draw_surf)

        self.ui.makeMovieBtn.clicked.connect(self.make_movie)
        self.movie_progress_signal.connect(self.ui.movieProgressBar.setValue)

        # Misc
        try:
            self._load_model(tmp_path)
        except:
            self.ui.statusbar.showMessage("Please choose a model.")
        else:
            self.ui.statusbar.showMessage("Using cached model.")

        self.show()

    # Panneau modèle

    def set_model(self, model):
        self.model = model

        self.ui.saveAllAtomsPositionCheckBox.setEnabled(True)
        self.ui.saveModelBtn.setEnabled(True)

        self.ui.gotoSimuBtn.setEnabled(True)
        self.ui.tab_simu.setEnabled(True)

        self.simulation = Simulation(self.model, prefer_gpu=self.ui.tryToUseGPUCheckBox.checkState())
        self.model_view = ModelView(self.simulation.model)

        for dp in self.displayed_properties:
            self.displayed_properties[dp].setText(1, str(self.model.__getattr__(dp)))

        self.update_simu_time()

        self.model_to_cache()

        self.ui.tab_processing.setEnabled(False)
        self.ui.currentTime.setText("0")
        self.ui.currentIteration.setText("0")
        self.ui.editModelBtn.setEnabled(True)
        self.ui.newSimuBtn.setEnabled(True)

        self.ui.statusbar.showMessage("Model loaded, simulation can begin.")

    def reset_model(self):
        self.set_model(self.model)

    def edit_model(self):
        self.emd = EditModelDialog(self)
        self.emd.show()

    def load_simulation(self):
        path, filter = QFileDialog.getOpenFileName(caption="Load model", filter=SIMULATION_FILE_FILTER_)
        if path:
            ds = self._load_model(path)
            with ds.open(ds.STATE_FCT, 'r') as IO:
                for key, item in IO.items():
                    self.simulation.state_fct[key] = item
            c_i = len(self.simulation.state_fct["T"])
            self.simulation.current_iter = c_i
            self.ui.currentIteration.setText(str(c_i))
            self.ui.currentTime.setText(str((c_i)*self.model.dt))
            self.enable_process_tab(True)
            t, T = self.simulation.state_fct["T_ramps"]
            if len(t)>1:
                self.simulation.set_T_ramps(t, T)
            t, Fx = self.simulation.state_fct["Fx_ramps"]
            if len(t)>1:
                self.simulation.set_Fx_ramps(t, Fx)
            t, Fy = self.simulation.state_fct["Fy_ramps"]
            if len(t)>1:
                self.simulation.set_Fy_ramps(t, Fy)
            sp = self.simulation.model.params["save_pos_history"]
            self.ui.groupBoxMovie.setEnabled(sp)
            self.ui.saveAllAtomsPositionCheckBox.setCheckState(sp)
            self.ui.saveAllAtomsPositionCheckBox.setEnabled(False)
            self.ui.statusbar.showMessage("Simulation history loaded.")

    def show_model(self):
        self.model_view.show() # bidon mais nécessaire pour que ça marche : on risque de redéfinir model donc model_view

    def create_model(self):
        self.cmd = CreateModelDialog(self)
        self.cmd.show()

    def load_model(self):
        path, filter = QFileDialog.getOpenFileName(caption="Load model", filter=MODEL_FILE_FILTER)
        if path:
            self._load_model(path)

    def _load_model(self, path):
        ds = DynState(path)
        model = Model()
        # position of particles
        with ds.open(ds.POS, 'r') as IO:
            model.pos = IO.load()
        # parameters
        with ds.open(ds.PAR) as IO:
            for key, item in IO.items():
                model.params[key] = item
        # velocity
        with ds.open(ds.VEL, 'r') as IO:
            model.v = IO.load()

        model._m()

        self.set_model(model)
        return ds

    def _save_model(self, m):
        path, filter = QFileDialog.getSaveFileName(caption="Save model", filter=MODEL_FILE_FILTER)
        print(filter)
        if path:
            path = self._correct_path(path, filter, [".zip", ".mdl"])
            try:
                shutil.rmtree(tmp1_path)
            except FileNotFoundError:
                pass
            ds = DynState(tmp1_path)
            ds.save_model(m)
            ds.to_zip(path)

    def _correct_path(self, path, filter, expected):
        for ext in expected:
            if not path.endswith(ext) and ext in filter:
                path += ext
                return path
        return path

    def save_model(self):
        self._save_model(self.model)

    def save_final_model(self):
        self._save_model(self.simulation.model)

    # Panneau simu

    def _model_to_cache(self, index):
        if (self.ui.tabWidget.currentWidget() is self.ui.tab_simu):
            self.model_to_cache()

    def model_to_cache(self):
        if self.model:
            self.ds = DynState(tmp_path)
            self.ds.save_model(self.model)

    def update_iters(self):
        try:
            t = _float(self.ui.simulationTimeLineEdit)
        except:
            self.update_simu_time()
        else:
            self.ui.iterationsSpinBox.setValue(int(t/self.simulation.model.dt))

    def update_simu_time(self, v=None):
        self.ui.simulationTimeLineEdit.setText(str(self.model.dt*self.ui.iterationsSpinBox.value()))
        self.ui.simuProgressBar.setMaximum(self.ui.iterationsSpinBox.value())
        self.progress_plt.setXRange(0, self.ui.iterationsSpinBox.value())

    def goto_simu(self):
        self.ui.tabWidget.setCurrentWidget(self.ui.tab_simu)

    def design_temperature_profile(self):
        queue = Queue(1)
        axis = (0, (self.ui.iterationsSpinBox.value()+self.simulation.current_iter)*self.model.dt, 0, 100)
        queue.put((axis, *self.simulation.T_ramps, "Time (s)", "Temperature (K)", 0))
        design_thread = Process(target=draggableLine.main, args=(queue,))
        design_thread.start()
        design_thread.join()
        x_data, y_data = queue.get()
        queue.close()
        if len(x_data)<2:
            qmb = QMessageBox()
            qmb.setText("This only works with at least two points.")
            qmb.exec()
            return
        self.simulation.set_T_ramps(x_data, y_data)

    def update_progress(self, v, new_t):
        c_i = v+1 - self.c_i
        self.t_deque.append(1/(new_t - self.last_t))
        self.last_t = new_t
        self.ui.simuProgressBar.setValue(c_i)
        if c_i==self.ui.iterationsSpinBox.value() or not c_i%10:
            self.progress_gr.setData(self.t_deque)
            self.ui.currentIteration.setText(str(v+1))
            self.ui.currentTime.setText(str((v+1)*self.model.dt))
            self.ui.ETA.setText(str(timedelta(seconds=int( (self.ui.iterationsSpinBox.value()/c_i - 1)*(new_t-self.simu_starttime)))))

        if self.save_pos:
            self.pos_IO.save(self.simulation.model.pos)

    def simulate(self):
        self.ui.simuBtn.setEnabled(False)
        self.ui.iterationsSpinBox.setEnabled(False)
        self.ui.simulationTimeLineEdit.setEnabled(False)
        self.ui.tryToUseGPUCheckBox.setEnabled(False)
        self.ui.temperature_groupBox.setEnabled(False)
        self.ui.saveAllAtomsPositionCheckBox.setEnabled(False)
        self.enable_process_tab(False)
        self.ui.simuProgressBar.setValue(0)
        self.last_t = time.perf_counter()
        self.t_deque.clear()
        self.t_deque.append(0)
        self.ui.statusbar.showMessage("Simulation is running...")

        self.save_pos = self.ui.saveAllAtomsPositionCheckBox.checkState()
        self.simulation.model.params["save_pos_history"] = self.save_pos
        if self.save_pos:
            mode = "a" if self.simulation.current_iter else "w"

            self.pos_IO = DynState(tmp_path).open(DynState.POS_H, mode=mode)
            self.pos_IO.__enter__()

        if len(self.simulation.T_ramps[0]):
            final_t = (self.simulation.current_iter + self.ui.iterationsSpinBox.value())*self.model.dt
            t, T = map(list,self.simulation.T_ramps)
            t.append(final_t)
            T.append(self.simulation.T_f(final_t))
            self.simulation.set_T_ramps(t, T)

        def run():
            # Pour continuer la simu précedente. On est obligés d'en créer une nouvelle pour des questions de scope.
            # On pourrait créer et conserver le thread une bonne fois pour toutes, pour que ce bricolage cesse.
            self.c_i = self.simulation.current_iter
            self.simulation = Simulation(simulation=self.simulation, prefer_gpu=self.ui.tryToUseGPUCheckBox.checkState())
            self.model_view = ModelView(self.simulation.model)
            self.simu_starttime = time.perf_counter()
            def up(s):
                self.updated_signal.emit(s.current_iter, time.perf_counter())
            self.simulation.iter(self.ui.iterationsSpinBox.value(), up)
            self.simu_thr.exit()

        def end():
            if self.save_pos:
                self.pos_IO.file.close()

            with DynState(tmp_path).open(DynState.STATE_FCT, mode="w") as ds:
                ds.from_dict(self.simulation.state_fct)

            self.enable_process_tab(True)
            self.ui.simuBtn.setEnabled(True)
            self.ui.iterationsSpinBox.setEnabled(True)
            self.ui.tryToUseGPUCheckBox.setEnabled(True)
            self.ui.simulationTimeLineEdit.setEnabled(True)
            self.ui.temperature_groupBox.setEnabled(True)
            self.ui.groupBoxMovie.setEnabled(self.save_pos)
            self.ui.statusbar.showMessage("Simulation complete. (Iterations : " + str(self.simulation.current_iter) + ")")

        self.simu_thr = QThread()
        self.simu_thr.run = run
        self.simu_thr.finished.connect(end)
        self.simu_thr.start()

    def enable_process_tab(self, b):
        self.ui.tab_processing.setEnabled(b)
        self.ui.gotoProcessBtn.setEnabled(b)

    def goto_process(self):
        self.ui.tabWidget.setCurrentWidget(self.ui.tab_processing)

    # Panneau process

    def reuse_model(self):
        self.set_model(self.simulation.model)
        self.ui.tabWidget.setCurrentWidget(self.ui.tab_model)

    def save_simu_history(self):
        path, filter = QFileDialog.getSaveFileName(caption="Save simulation history", filter=SIMULATION_FILE_FILTER_)
        if path:
            path = self._correct_path(path, filter, [".zip", ".mds"])
            #shutil.rmtree('./data/tmp1')
            ds = DynState(tmp_path)
            if not self.ui.saveAllAtomsPositionCheckBox.checkState():
                try:
                    os.remove(tmp_path+'/'+DynState.POS_H)
                except:
                    pass
            ds.save_model(self.simulation.model)
            ds.to_zip(path)

    def export_to_csv(self):
        path, filter = QFileDialog.getSaveFileName(caption="Export to CSV", filter="CSV file (*.csv)")
        if path:
            path = self._correct_path(path, filter, [".csv"])
            with open(path, "w", newline='') as csvfile:
                csvwriter = csv.writer(csvfile)
                csvwriter.writerow(self.temporal_variables.keys())
                csvwriter.writerows(zip(*(self.simulation.state_fct[s[0]] for s in self.temporal_variables.values())))

    def line_graph(self):
        ords = [i.text() for i in self.temp_variables_w if i.checkState()]
        if not len(ords):
            return

        absc = self.ui.lineComboW.currentText()

        def values(s):
            return self.simulation.state_fct[self.temporal_variables[s][0]]

        def label(s):
            if len(self.temporal_variables[s])>1:
                return s + " (" + self.temporal_variables[s][1] + ")"
            return s

        def dimension(s):
            if "energy" in s:
                return "Energy (J)"
            elif "Temperature" in s:
                return "Temperature (K)"
            else:
                return label(s)

        dimensions = list(set(dimension(s) for s in ords))
        axis = dict()

        visu.plt.ioff()
        fig = visu.plt.figure()

        gr_opts = dict()
        for o in self._1d_options:
            gr_opts[o] = self._1d_options[o].checkState()

        host = HostAxes(fig, [0.15, 0.1, 0.75-(0.04*len(dimensions)), 0.8])
        host.set_xlabel(label(absc))
        host.set_ylabel(dimensions[0])
        if gr_opts["log x"]:
            host.semilogx()
        if gr_opts["log y"]:
            host.semilogy()
        if gr_opts["grid"]:
            host.grid(True, which="minor", linestyle="--")
            host.grid(True, which="major")

        if len(dimensions)>1:
            host.axis["right"].set_visible(False)

        axis[dimensions[0]] = host

        for i, dim in enumerate(dimensions[1:]):
            par = ParasiteAxes(host, sharex=host)
            host.parasites.append(par)
            par.axis["right"] = par.get_grid_helper().new_fixed_axis(loc="right", axes=par, offset=(55*i, 0))
            par.axis["right"].set_visible(True)
            par.set_ylabel(dim)
            par.axis["right"].major_ticklabels.set_visible(True)
            par.axis["right"].label.set_visible(True)
            axis[dim] = par
            if gr_opts["log x"]:
                par.semilogx()
            if gr_opts["log y"]:
                par.semilogy()

        for i in ords:
            axis[dimension(i)].plot(values(absc), values(i), label=label(i))

        fig.add_axes(host)
        host.legend()
        visu.plt.show()

    def PDF(self):
        """Pair Distribution Function"""
        self.old_status = self.ui.statusbar.currentMessage()
        self.ui.statusbar.showMessage("Computing Pair Distribution Function...")

        visu.plt.ioff()
        d = PDF(self.simulation.model.pos, self.ui.PDFNSpinBox.value(), self.ui.PDFDistSpinBox.value()*max(self.model.rcut_a, self.model.rcut_b, self.model.rcut_ab), 100)
        visu.plt.figure()
        visu.plt.plot(*d)
        visu.plt.xlabel("Distance (m)")
        visu.plt.show()

        self.ui.statusbar.showMessage(self.old_status)

    def draw_surf(self):
        self.old_status = self.ui.statusbar.currentMessage()

        gr_opts = dict()
        for o in self._2d_options:
            gr_opts[o] = self._2d_options[o].checkState()

        visu.plt.figure()

        if gr_opts["density map"]:
            self.ui.statusbar.showMessage("Computing density map...")
            visu.plot_densityf(self.simulation.model, 50)

        if gr_opts["deformation (compression)"]:
            self.ui.statusbar.showMessage("Computing deformation...")
            visu.deformation_volume(self.model, self.simulation.model, self.ui.deformationDistSpinBox.value()*max(self.model.rcut_a, self.model.rcut_b, self.model.rcut_ab))

        if gr_opts["deformation (shear)"]:
            self.ui.statusbar.showMessage("Computing deformation...")
            visu.deformation_dev(self.model, self.simulation.model, self.ui.deformationDistSpinBox.value()*max(self.model.rcut_a, self.model.rcut_b, self.model.rcut_ab))

        if gr_opts["particles"]:
            visu.plot_particles(self.simulation.model)

        visu.plt.ylim(self.model.y_lim_inf, self.model.y_lim_sup)
        visu.plt.xlim(self.model.x_lim_inf, self.model.x_lim_sup)

        visu.plt.show()

        self.ui.statusbar.showMessage(self.old_status)

    def make_movie(self):
        path, filter = QFileDialog.getSaveFileName(caption="Make movie", filter="Video file (*.mp4)")
        if path:
            path = self._correct_path(path, filter, [".mp4"])
            self.ui.tab_model.setEnabled(False)
            self.ui.tab_simu.setEnabled(False)
            self.ui.groupBoxMovie.setEnabled(False)
            self.ui.movieProgressBar.setMaximum(self.simulation.current_iter)

            def run():
                def up(k):
                    self.movie_progress_signal.emit(k)

                visu.make_movie(self.simulation, DynState(tmp_path), path, self.ui.stepsByFrameSpinBox.value(),
                                self.ui.FPSSpinBox.value(), callback=up)

            def end():
                self.ui.tab_model.setEnabled(True)
                self.ui.tab_simu.setEnabled(True)
                self.ui.groupBoxMovie.setEnabled(True)
                self.ui.movieProgressBar.setValue(self.simulation.current_iter)

            self.render_thr = QThread()
            self.render_thr.run = run
            self.render_thr.finished.connect(end)
            self.render_thr.start()


"""Tests for manipulators.

"""

import datreant as dtr
from datreant import discover


def test_discover(tmpdir):
    with tmpdir.as_cwd():

        ghosts = ('inky', 'blinky', 'pinky', 'clyde')

        for name in ghosts:
            dtr.Treant(
                    'a/very/deep/directory/structure/that/just/keeps/going/' +
                    name)

        b = discover('.')

        assert len(b) == 4

        for name in ghosts:
            assert name in b.names


def test_discover_depth(tmpdir):
    """Check that using `depth` parameter gives expected result."""
    with tmpdir.as_cwd():

        ghosts = ('something/inky',
                  'something/else/blinky',
                  'pinky',
                  'something/clyde')

        for name in ghosts:
            dtr.Treant(name)

        assert len(discover('.', depth=0)) == 0
        assert len(discover('pinky', depth=0)) == 1

        assert len(discover('.', depth=1)) == 1
        assert len(discover('.', depth=2)) == 3
        assert len(discover('.', depth=3)) == 4


def test_discover_treantdepth(tmpdir):
    """Check that using `treantdepth` parameter gives expected result."""
    with tmpdir.as_cwd():

        ghosts = ('inky',
                  'inky/blinky',
                  'pinky',
                  'inky/blinky/clyde')

        for name in ghosts:
            dtr.Treant(name)

        assert len(discover('.', treantdepth=0)) == 2
        assert len(discover('pinky', treantdepth=0)) == 1
        assert len(discover('inky', treantdepth=0)) == 1

        assert len(discover('.', treantdepth=1)) == 3
        assert len(discover('.', treantdepth=2)) == 4
        assert len(discover('inky', treantdepth=1)) == 2
        assert len(discover('inky/blinky', treantdepth=1)) == 2

        assert len(discover('inky/blinky', treantdepth=1)) == 2


def test_discover_depth_treantdepth(tmpdir):
    """Check that using `treantdepth` and `depth` parameters together gives
        expected result.
    """
    with tmpdir.as_cwd():

        ghosts = ('inky',
                  'inky/blinky',
                  'pinky',
                  'inky/blinky/nothing/clyde')

        for name in ghosts:
            dtr.Treant(name)

        assert len(discover('.', treantdepth=0, depth=0)) == 0
        assert len(discover('.', treantdepth=0, depth=1)) == 2
        assert len(discover('pinky', treantdepth=0, depth=0)) == 1
        assert len(discover('inky', treantdepth=0, depth=2)) == 1

        assert len(discover('.', treantdepth=1, depth=1)) == 2
        assert len(discover('inky', treantdepth=1, depth=1)) == 2
        assert len(discover('inky', treantdepth=1, depth=0)) == 1

        assert len(discover('inky', treantdepth=2)) == 3
        assert len(discover('inky', treantdepth=2, depth=2)) == 2
        assert len(discover('inky', treantdepth=2, depth=3)) == 3



import matplotlib
matplotlib.use('Qt5Agg')
import matplotlib.pyplot as plt

class ModelView():
    def __init__(self, model):
        self.model = model

        plt.ioff()

        plt.pause(1e-6)

    def update(self):
        self.plot_a.set_data(*self.model.pos[:self.model.n_a,:].T)
        self.plot_b.set_data(*self.model.pos[self.model.n_a:,:].T)
        plt.pause(1e-6)

    def show(self):
        plt.clf()
        plt.axis("scaled")
        plt.ylim(self.model.y_lim_inf, self.model.y_lim_sup)
        plt.xlim(self.model.x_lim_inf, self.model.x_lim_sup)
        self.plot_a, = plt.plot(*self.model.pos[:self.model.n_a,:].T, "ro", markersize=1)
        self.plot_b, = plt.plot(*self.model.pos[self.model.n_a:,:].T, "bo", markersize=1)
        plt.pause(1e-3)
        #return self


import pytest
from glob import glob
from os import path
import json

import datreant as dtr
from datreant.scripts.datreant_07to1 import convert


@pytest.fixture
def old_treant():
    return {
        "categories": {
            "totaltime": 2336232.0,
            "selection": " name CA ",
            "reference": "../../ave_struc_all_ca.pdb",
            "time_steps": 8333,
            "sim-path": "box1/production/traj-2/",
            "protein": "myoglobin",
            "dt": 6.0
        },
        "tags": ["foo", "BAR"]
    }


def test_convert(old_treant, tmpdir):
    treant_folder = str(tmpdir)
    with open(path.join(treant_folder, 'Treant-uuid.json'), 'w') as fh:
        json.dump(old_treant, fh)

    convert(treant_folder)
    treant = dtr.Treant(treant_folder)

    assert treant.tags == old_treant['tags']
    assert treant.categories == old_treant['categories']


def test_convert_no_treant(tmpdir):
    treant_folder = str(tmpdir)
    with pytest.warns(UserWarning, match="No treant found.*"):
        convert(treant_folder)


def test_convert_to_many_treants(tmpdir, old_treant):
    treant_folder = str(tmpdir)
    with open(path.join(treant_folder, 'Treant-uuid.json'), 'w') as fh:
        json.dump(old_treant, fh)

    with open(path.join(treant_folder, 'Treant-uuid2.json'), 'w') as fh:
        json.dump(old_treant, fh)

    with pytest.warns(UserWarning, match="Multiple treants found.*"):
        convert(treant_folder)



from PyQt5.QtWidgets import QWidget
from PyQt5.QtCore import pyqtSignal

from .qt.species_params import Ui_species_params
from ..data.atom_preset import atoms

from ._conv import _float

class species_params(QWidget):
    editingFinished = pyqtSignal()

    def __init__(self,*args):
        super().__init__(*args)
        self.ui = Ui_species_params()
        self.ui.setupUi(self)

        for atom in atoms:
            self.ui.presetComboBox.addItem(atom)

        self.ui.presetComboBox.currentTextChanged.connect(self.set_preset)

        self.ui.sigmaMLineEdit.editingFinished.connect(self._editing_finished)
        self.ui.epsilonJLineEdit.editingFinished.connect(self._editing_finished)
        self.ui.massKgLineEdit.editingFinished.connect(self._editing_finished)

        self.show()

    def set_preset(self, a):
        if a in atoms:
            self.ui.nameLineEdit.setText(a)
            self.set_values(atoms[a])
            self._editing_finished()

    def check_values(self):
        try:
            self.get_values()
        except:
            return False
        return True

    def set_values(self, v, force = True):
        ws = (self.ui.epsilonJLineEdit, self.ui.sigmaMLineEdit, self.ui.massKgLineEdit)
        for w, av in zip(ws, v):
            lf = False
            try:
                float(w.text())
            except:
                lf = True
            if lf or force:
                w.setText(str(av))

    def get_values(self):
        wid = (
            self.ui.epsilonJLineEdit,
            self.ui.sigmaMLineEdit,
            self.ui.massKgLineEdit,
        )
        return tuple(map(_float,wid))

    def _editing_finished(self):
        self.editingFinished.emit()


"""
Treants: the organizational units for :mod:`datreant`.

"""
import os
import functools
import six
from pathlib2 import Path

from .collections import Bundle
from .trees import Tree
from .names import TREANTDIR_NAME
from .util import makedirs
from .metadata import Tags, Categories
from .exceptions import NotATreantError


@functools.total_ordering
class Treant(Tree):
    """The Treant: a discoverable Tree with metadata.

    `treant` should be the directory of a new or existing Treant.
    An existing Treant will be used if a `.datreant` directory is found inside.
    If no `.datreant` directory is found inside, a new Treant will be created.

    A Tree object may also be used in the same way as a directory string.

    Parameters
    ----------
    treant : str or Tree
        Base directory of a new or existing Treant; may also be a Tree object
    categories : dict
        dictionary with user-defined keys and values; used to give
        Treants distinguishing characteristics
    tags : list
        list with user-defined strings; like categories, but useful for
        adding many distinguishing descriptors
    """

    def __init__(self, treant, categories=None, tags=None):
        # if given a Tree, get path out of it
        if isinstance(treant, Tree):
            treant = treant.abspath

        super(Treant, self).__init__(treant)

        # attach metadata objects
        self._tags = Tags(self)
        self._categories = Categories(self)

        # make treantdir
        self._make_treantdir()

        if tags:
            self.tags.add(tags)
        if categories:
            self.categories.add(categories)

    def _make_treantdir(self):
        abspath = super(Treant, self).__getattribute__('_path').absolute()
        treantdir = abspath / TREANTDIR_NAME

        if not treantdir.exists():
            # build datreant dir; stop if we hit a permissions error
            try:
                makedirs(treantdir, exist_ok=True)
            except OSError as e:
                if e.errno == 13:
                    raise OSError(13, "Permission denied; " +
                                  "cannot create '{}'".format(treantdir))
                else:
                    raise

    def __repr__(self):
        return "<Treant: '{}'>".format(self.name)

    def __getstate__(self):
        return self.abspath

    def __setstate__(self, state):
        self.__init__(state)

    def __hash__(self):
        return hash(self.abspath)

    def __eq__(self, other):
        try:
            return self.abspath == other.abspath
        except AttributeError:
            return NotImplemented

    def __lt__(self, other):
        try:
            return self.name < other.name
        except AttributeError:
            return NotImplemented

    def __add__(self, other):
        """Addition of treants with collections or treants yields Bundle.

        """
        if isinstance(other, (Treant, Bundle)):
            return Bundle(self, other)
        else:
            raise TypeError("Operands must be Treants or Bundles.")

    @property
    def name(self):
        """The name of the Treant.

        """
        return super(Treant, self).name

    @property
    def _treantdir(self):
        return os.path.join(self.abspath, TREANTDIR_NAME)

    @property
    def tags(self):
        try:
            return self._tags
        except FileNotFoundError:
            raise NotATreantError("This Treant no longer has a '.datreant'"
                                  " directory.")

    @tags.setter
    def tags(self, value):
        try:
            if isinstance(value, (Tags, list, set)):
                val = list(value)
                self.tags.clear()
                self.tags.add(val)
            else:
                raise TypeError("Can only set with tags, a list, or set")
        except FileNotFoundError:
            raise NotATreantError("This Treant no longer has a '.datreant'"
                                  " directory.")

    @property
    def categories(self):
        try:
            return self._categories
        except FileNotFoundError:
            raise NotATreantError("This Treant no longer has a '.datreant'"
                                  " directory.")

    @categories.setter
    def categories(self, value):
        try:
            if isinstance(value, (Categories, dict)):
                val = dict(value)
                self.categories.clear()
                self.categories.add(val)
            else:
                raise TypeError("Can only set with categories or dict")
        except FileNotFoundError:
            raise NotATreantError("This Treant no longer has a '.datreant'"
                                  " directory.")
